# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Driver module."""

try:
    from threading import Thread

except ImportError:
    from dummy_threading import Thread

from .transaction import State, Transaction

from ..request.consts import FuncName
from ..request.crud.create import Create
from ..request.crud.delete import Delete
from ..request.crud.read import Read
from ..request.crud.update import Update

__all__ = ['Driver']


DEFAULT_ASYNC = False  #: default async value.


class Driver(object):
    """In charge of accessing to data from a transaction."""

    name = None  #: driver name.
    #: supported function names.
    supportedfunctions = [item.value for item in FuncName.__members__.values()]

    def __init__(self, name=None, *args, **kwargs):

        super(Driver, self).__init__(*args, **kwargs)

        if name is not None:
            self.name = name

    def open(self, ctx=None, autocommit=False, cruds=None):
        """open a new transaction.

        :param Context ctx: execution context.
        :param bool autocommit: transaction autocommit property. False by
            default.
        :rtype: Transaction.
        """

        return Transaction(
            driver=self, ctx=ctx, autocommit=autocommit, cruds=cruds
        )

    def process(
            self, transaction, async=DEFAULT_ASYNC, callback=None,
            **kwargs
    ):
        """Process input transaction and crud element.

        :param Transaction transaction: transaction to process.
        :param bool async: if True (default False), execute input crud in a
            separated thread.
        :param Callable callback: callable function which takes in parameter
            the function result and kwargs. Commonly used with async equals
            True.
        :param dict kwargs: additional parameters specific to the driver.
        :return: transaction or thread if async.
        :rtype: Transaction or Thread
        """
        result = transaction

        def process():

            if transaction.dparams is not None:
                kwargs.update(transaction.dparams)

            result = self._process(
                transaction=transaction, async=async, **kwargs
            )

            if callback is not None:
                callback(result, async=async, **kwargs)

            return result

        if transaction.state is State.COMMITTING:

            if async:
                result = Thread(target=process)
                result.start()

            else:
                result = process()

        return result

    def _process(self, transaction, **kwargs):
        """Generic method to override in order to crud input data related to
        transaction and kwargs.

        :param Transaction transaction: transaction to process.
        :param dict kwargs: additional parameters specific to the driver.
        :return: transaction.
        :rtype: Transaction
        """

        raise NotImplementedError()

    def __repr__(self):

        return 'Driver({0})'.format(self.name)

    def __str__(self):

        return repr(self)

    def __getitem__(self, key):

        return self.read(key).ctx[key]

    def __setitem__(self, key, value):

        return self.update(name=key, **value)

    def __delitem__(self, key):

        return self.delete(key)

    def _getcrud(self, cls, **kwargs):

        crud = cls(**kwargs)

        return self.open(cruds=[crud]).commit()

    def create(self, **kwargs):
        """Create creation."""

        return self._getcrud(cls=Create, **kwargs)

    def read(self, **kwargs):
        """Read input expressions.

        :param tuple select: selection fields.
        :param dict kwargs: additional selection parameters (limit, etc.).
        :rtype: Cursor"""

        return self._getcrud(cls=Read, **kwargs)

    def update(self, **kwargs):
        """Apply input updates.

        :param tuple update: updates to apply."""

        return self._getcrud(cls=Update, **kwargs)

    def delete(self, **kwargs):
        """Delete input deletes.

        :param tuple names: model name to delete.
        :return: number of deleted deletes."""

        return self._getcrud(cls=Delete, **kwargs)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Python driver module."""

from datetime import datetime

from functools import wraps

from hashlib import md5

from operator import (
    abs as abs_, add, concat, contains, countOf, delitem, eq,
    floordiv, ge, getitem, gt, iadd, iand, iconcat, ifloordiv,
    ilshift, imod, imul, indexOf, invert,

    ior, ipow, irshift, is_, is_not, isub,
    itruediv, ixor,
    le, lshift, lt, mod, mul, ne, neg, not_, pow as pow_, rshift,
    setitem, sub, truediv, truth, xor,
)

from random import random

from re import match

from b3j0f.utils.path import lookup

from six import iteritems

from soundex import getInstance


from .base import Driver

from .ctx import Context
from .transaction import State

from .utils import FunctionChooser
from ..request.base import BaseElement
from ..request.consts import CONDITIONS, FuncName
from ..request.crud.base import CRUDElement
from ..request.crud.create import Create
from ..request.crud.delete import Delete
from ..request.crud.read import Read
from ..request.crud.update import Update
from ..request.expr import Expression, Function

__all__ = [
    'PyDriver', 'processcrud', 'processquery',
    'processcreate', 'processread', 'processupdate', 'processdelete',
    'FunctionChooser'
]

VERSION = '0.1'  #: python driver version.

soundex = getInstance().soundex

DTF = '%Y-%m-%d %H:%M:%S'  #: date time format


def processcreate(items, create, ctx=None, **kwargs):
    """Apply input Create element to items.

    :param list items: items to process with input Create.
    :param Create create: data to add to input items.
    :return: created item.
    :rtype: list
    """
    return _GLOBALPYDRIVER.processcreate(
        create=create, ctx=ctx, items=items, **kwargs
    )


def processread(items, read, ctx=None, **kwargs):
    """Return application of input Read to items.

    :param list items: items to read.
    :param Read read: read resource to apply on items.
    :return: read list.
    :rtype: list
    """
    return _GLOBALPYDRIVER.processread(
        read=read, ctx=ctx, items=items, **kwargs
    )


def processupdate(items, update, ctx=None, **kwargs):

    return _GLOBALPYDRIVER.processupdate(
        items=items, update=update, ctx=ctx, **kwargs
    )


def processdelete(items, delete, ctx=None, **kwargs):

    return _GLOBALPYDRIVER.processdelete(
        items=items, delete=delete, ctx=ctx, **kwargs
    )


def processcrud(items, crud, ctx=None, **kwargs):

    return _GLOBALPYDRIVER.processcrud(
        items=items, crud=crud, ctx=ctx, **kwargs
    )


def exists(query, item, name, fparams, ctx):

    result = True

    names = name.split('.')

    val = item

    for name in names:
        if isinstance(val, dict):

            if name in val:
                val = val[name]
                continue

        elif hasattr(val, name):
            val = getattr(val, name)
            continue

        result = False
        break

    return result


def isnull(query, item, name, fparams, ctx):

    try:
        val = getsubitem(item, name, error=True)

    except KeyError:
        return False

    else:
        return val is None


def _namedelt(operator):

    def result(query, item, name, params, ctx):

        subitem = getsubitem(item, name)

        return operator(subitem, *params[1:])

    return result


def all_(query, item, name, params, ctx):

    items = params[2]

    result = len(items) > 0

    operator = query.params[1]

    if isinstance(operator, Expression):
        operator = operator.name

    func = _OPERATORS_BY_NAME[operator]

    for _item in items:

        function = Function(operator)(query.params[0], _item)

        fparams = [[item], _item]

        if not func(function=function, ctx=ctx, params=fparams):
            result = False
            break

    return result


def any_(query, item, name, params, ctx):

    items = params[2]

    result = False

    operator = query.params[1]

    if isinstance(operator, Expression):
        operator = operator.name

    func = _OPERATORS_BY_NAME[operator]

    for _item in items:

        function = Function(operator)(query.params[0], _item)

        fparams = [[item], _item]

        if func(function=function, ctx=ctx, params=fparams):
            result = True
            break

    return result


_OPERATORS_BY_NAME = {
    FuncName.AND.value: lambda function, params, ctx: reduce(add, params),
    FuncName.OR.value: lambda function, params, ctx: reduce(add, params),
    FuncName.LT.value: _namedelt(lt),
    FuncName.LE.value: _namedelt(le),
    FuncName.EQ.value: _namedelt(eq),
    FuncName.NE.value: _namedelt(ne),
    FuncName.GE.value: _namedelt(ge),
    FuncName.GT.value: _namedelt(gt),
    FuncName.NOT.value: lambda function, params, ctx: not_(params[0]),
    FuncName.TRUTH.value: truth,
    FuncName.IS.value: _namedelt(is_),
    FuncName.ISNOT.value: _namedelt(is_not),
    FuncName.ABS.value: abs_,
    FuncName.ADD.value: add,
    FuncName.FLOORDIV.value: floordiv,
    FuncName.DIV.value: truediv,
    FuncName.INDEX.value: indexOf,
    FuncName.INVERT.value: invert,
    FuncName.MOD.value: mod,
    FuncName.LIKE.value: _namedelt(lambda val, msg: match(msg, val)),
    FuncName.MUL.value: mul,
    FuncName.NEG.value: neg,
    FuncName.POW.value: pow_,
    FuncName.RSHIFT.value: rshift,
    FuncName.LSHIFT.value: lshift,
    FuncName.SUB.value: sub,
    FuncName.XOR.value: xor,
    FuncName.CONCAT.value: concat,
    FuncName.COUNTOF.value: countOf,
    FuncName.REPEAT.value: mul,
    FuncName.INCLUDE.value: contains,
    FuncName.IADD.value: iadd,
    FuncName.IAND.value: iand,
    FuncName.IOR.value: ior,
    FuncName.IXOR.value: ixor,
    FuncName.GETITEM.value: getitem,
    FuncName.SETITEM.value: setitem,
    FuncName.DELITEM.value: delitem,
    FuncName.GETSLICE.value: getitem,
    FuncName.SETSLICE.value: setitem,
    FuncName.DELSLICE.value: delitem,
    FuncName.ICONCAT.value: iconcat,
    FuncName.IDIV.value: itruediv,
    FuncName.IFLOORDIV.value: ifloordiv,
    FuncName.ILSHIFT.value: ilshift,
    FuncName.IMOD.value: imod,
    FuncName.IMUL.value: imul,
    FuncName.IPOW.value: ipow,
    FuncName.IREPEAT.value: imul,
    FuncName.IRSHIFT.value: irshift,
    FuncName.ISUB.value: isub,
    FuncName.COUNT.value: len,
    FuncName.LENGTH.value: len,
    FuncName.AVG.value: lambda v: sum(v) / (len(v) or 1),
    FuncName.MEAN.value: lambda v: sum(v) / (len(v) or 1),
    FuncName.MAX.value: max,
    FuncName.MIN.value: min,
    FuncName.SUM.value: sum,
    FuncName.EXISTS.value: exists,
    FuncName.NEXISTS.value:
        lambda query, item, name, *_: not exists(query, item, name, *_),
    FuncName.ISNULL.value: isnull,
    FuncName.BETWEEN.value: _namedelt(lambda val, inf, sup: inf <= val <= sup),
    FuncName.IN.value: _namedelt(lambda x, y: contains(y, x)),
    FuncName.HAVING.value: lambda query, item, name, fparams, ctx: fparams[0],
    FuncName.UNION.value: lambda seq1, seq2: set(seq1) + set(seq2),
    FuncName.INTERSECT.value: lambda seq1, seq2: set(seq1) & set(seq2),
    FuncName.ALL.value: all_,
    FuncName.ANY.value: any_,
    FuncName.SOME.value: any_,
    FuncName.VERSION.value: VERSION,
    FuncName.CONCAT.value: str.__add__,
    FuncName.ICONCAT.value: str.__add__,
    FuncName.REPLACE.value: str.replace,
    FuncName.SOUNDEX.value: soundex,
    FuncName.SUBSTRING.value: lambda data, start, end=None: data[start:end],
    FuncName.LEFT.value: lambda data, count: str[:-count],
    FuncName.RIGHT.value: lambda data, count: str[count:],
    FuncName.REVERSE.value: reversed,
    FuncName.TRIM.value: str.strip,
    FuncName.LTRIM.value: str.lstrip,
    FuncName.RTRIM.value: str.rstrip,
    FuncName.LPAD.value: str.ljust,
    FuncName.RPAD.value: str.rjust,
    FuncName.UPPER.value: str.upper,
    FuncName.LOWER.value: str.lower,
    FuncName.UCASE.value: str.upper,
    FuncName.LCASE.value: str.lower,
    FuncName.LOCATE.value: lambda val, data, *args: str.find(
        data, val, *args
    ) + 1,
    FuncName.INSTR.value: lambda val, data, *args: str.find(
        data, val, *args
    ) + 1,
    FuncName.RAND.value: random,
    FuncName.ROUND.value: round,
    FuncName.MD5.value: lambda data: md5(data).digest(),
    FuncName.NOW.value: lambda: datetime.now().strftime(DTF),
    FuncName.SEC_TO_TIME.value:
        lambda date: datetime.fromtimestamp(date).strftime(DTF),
    FuncName.DATEDIFF.value:
        lambda date1, date2:
            datetime.strpformat(date1, DTF) - datetime.strpformat(date2, DTF),
    FuncName.MONTH.value:
        lambda date=None: datetime.strpformat(date, DTF).month,
    FuncName.YEAR.value:
        lambda date=None: datetime.strpformat(date, DTF).year,
}


def processquery(query, items, ctx=None, **kwargs):

    return _GLOBALPYDRIVER.processquery(
        query=query, ctx=ctx, items=items, **kwargs
    )


def getsubitem(item, name, error=False):

    names = name.split('.')

    result = item

    for name in names:
        try:
            result = result[name]

        except KeyError:
            if error:
                raise

            else:
                result = None
                break

    return result


def condoperator(operator):

    @wraps(operator)
    def result(function, params, ctx):

        if params:

            name = function.params[0].name

            result = [
                item for item in params[0]
                if operator(function, item, name, params, ctx)
            ]

            params[0][:] = result

        else:
            result = function

        return result

    return result


for condition in CONDITIONS:
    _condoperator = condoperator(_OPERATORS_BY_NAME[condition])
    _condoperator.__name__ = condition
    _OPERATORS_BY_NAME[condition] = _condoperator


class PyFunctionChooser(object):

    def get(self, name):

        if name in _OPERATORS_BY_NAME:
            return _OPERATORS_BY_NAME[name]

        else:
            try:
                return lookup(name)

            except ImportError:
                pass


class PyDriver(Driver):
    """In charge of accessing data from a list of dictionaries or objects."""

    name = 'py'  # driver name

    def __init__(
            self, items=None, funcchooser=PyFunctionChooser(), *args, **kwargs
    ):
        """
        :param list items: list of data. Data are dictionaries. Default is [].
        """

        super(PyDriver, self).__init__(*args, **kwargs)

        self.items = [] if items is None else items
        self.funcchooser = funcchooser

    def _process(self, transaction, **kwargs):

        result = transaction

        if transaction.state is State.COMMITTING:
            for crud in transaction.cruds:
                self.processcrud(ctx=transaction.ctx, crud=crud, **kwargs)

        return result

    def processcrud(self, crud, ctx=None, **kwargs):
        """Apply the right rule.

        :param CRUDElement crud: crud rule to apply.
        :param Context ctx: context where to store the processing result.
            Default is None.

        :rtype: list
        :return: list
        """
        result = kwargs.setdefault('items', self.items)

        if ctx is None:
            ctx = Context()

        if crud in ctx:  # is crud already calculated
            result = ctx[crud]

        else:
            if isinstance(crud, Create):
                result = self.processcreate(create=crud, ctx=ctx, **kwargs)

            elif isinstance(crud, Read):
                result = self.processread(read=crud, ctx=ctx, **kwargs)

            elif isinstance(crud, Update):
                result = self.processupdate(update=crud, ctx=ctx, **kwargs)

            elif isinstance(crud, Delete):
                result = self.processdelete(delete=crud, ctx=ctx, **kwargs)

            else:
                raise TypeError('{0} is not a crud element.'.format(crud))

            ctx[crud] = result

        return result

    def processquery(self, query, ctx=None, **kwargs):
        """Process input query related to items and ctx."""
        result = query

        if ctx is None:
            ctx = Context()

        if query in ctx:  # is query calculated already ?
            result = ctx[query]

        elif isinstance(query, BaseElement):

            if isinstance(query, Function):

                result = self.processfunction(
                    function=query, ctx=ctx, **kwargs
                )

            elif isinstance(query, Expression):

                result = self.processexpr(expr=query, ctx=ctx, **kwargs)

            elif isinstance(query, CRUDElement):

                result = self.processcrud(crud=query, ctx=ctx, **kwargs)

            ctx[query] = result

        return result

    def processfunction(self, function, ctx=None, **kwargs):

        items = kwargs.pop('items', self.items)

        func = self.getfunction(function=function, ctx=ctx)

        isor = function.name == FuncName.OR.value
        isand = function.name == FuncName.AND.value

        params = []

        fitems = presult = items
        fctx = ctx

        if isor:
            result = []

        for param in function.params:

            if isor:
                fitems = list(items)
                fctx = Context(ctx)

            elif isand:
                fitems = presult

            presult = self.processquery(
                query=param, ctx=fctx, items=fitems, **kwargs
            )

            params.append(presult)

            if isand:
                result = presult

            elif isor:
                result += presult

        if func is None:

            raise NotImplementedError(
                'Function {0} is not implented by {1}'.format(
                    function, self
                )
            )

        elif function.name not in [FuncName.AND.value, FuncName.OR.value]:

            result = func(function=function, ctx=ctx, params=params, **kwargs)

        ctx[function] = result

        return result

    def processexpr(self, expr, ctx=None, **kwargs):

        result = []

        if ctx is None:
            ctx = Context()

        items = kwargs.setdefault('items', self.items)

        for item in items:
            try:
                getsubitem(item, expr.name, error=True)

            except KeyError:
                pass

            else:
                result.append(item)

        if not result:  # if not item match, try to lookup the expression
            try:
                result = lookup(expr.name)

            except ImportError:
                items[:] = result

        else:
            items[:] = result

        ctx[expr] = result

        return result

    def processcreate(self, create, ctx=None, **kwargs):
        """Apply input Create element to items.

        :param Create create: data to add to input items.
        :return: created item.
        :rtype: list
        """
        result = kwargs.setdefault('items', self.items)

        if ctx is None:
            ctx = Context()

        if create.query is not None:
            result = self.processquery(query=create.query, ctx=ctx, **kwargs)

        values = create.values

        for key, query in iteritems(create.values):

            value = self.processquery(query=query, ctx=ctx, **kwargs)
            values[key] = value

        result.append(values)

        return result

    def processread(self, read, ctx=None, **kwargs):
        """Return application of input Read to items.

        :param list items: items to read.
        :param Read read: read resource to apply on items.
        :return: read list.
        :rtype: list
        """
        items = kwargs.setdefault('items', self.items)

        if ctx is None:
            ctx = Context()

        if read.query is not None:
            self.processquery(query=read.query, ctx=ctx, **kwargs)

        result = list(items)

        if read.select():
            result = []
            for item in list(items):
                fitem = {}
                for sel in read.select():
                    if sel in item:
                        fitem[sel] = item[sel]
                result.append(fitem)

        if read.offset():
            result = result[read.offset():]

        if read.limit():
            result = result[:read.limit()]

        if read.orderby():
            for orderby in read.orderby():
                result.sort(key=lambda item: item.get(orderby))

        if read.groupby():
            groupby = read.groupby()

            groupedvalues = {}

            for item in result:

                if groupby in item:

                    subitems = groupedvalues.setdefault(item[groupby], {})

                    for key, value in iteritems(item):

                        if key != groupby:
                            subitems.setdefault(key, []).append(value)

            result = []

            for key, value in iteritems(groupedvalues):

                value[groupby] = key
                result.append(value)

        if read.join() not in ('FULL', None):
            raise NotImplementedError(
                'read function does not support join {0}'.format(
                    read.join()
                )
            )

        return result

    def processupdate(self, update, ctx=None, **kwargs):
        """Apply update to items.

        :param Update update: update rule.
        :return: updated items.
        :rtype: list
        """
        if ctx is None:
            ctx = Context()

        result = kwargs.setdefault('items', self.items)

        if update.query is not None:
            result = self.processquery(query=update.query, ctx=ctx, **kwargs)

        values = {}

        for key in list(update.values):

            value = self.processquery(
                query=update.values[key], ctx=ctx, **kwargs
            )
            values[key] = value

        for item in result:
            for name, value in iteritems(values):

                if callable(value):
                    value(name, item)

                else:
                    item[name] = value

        return result

    def processdelete(self, delete, ctx=None, **kwargs):
        """Apply deletion rule to items.

        :param list items: items to modify.
        :param Delete delete: deletion rule.
        :rtype: list
        :return: modified/deleted items.
        """
        result = kwargs.setdefault('items', self.items)

        if ctx is None:
            ctx = Context()

        if delete.query is not None:
            result = self.processquery(query=delete.query, ctx=ctx, **kwargs)

        if delete.names():

            names = [
                self.processquery(query=name, ctx=ctx, **kwargs)
                for name in delete.names()
            ]

            for name in names:
                for item in result:
                    if name in item:
                        del item[name]

        else:
            result[:] = []

        return result

    def getfunction(self, function, ctx=None):

        return ctx.get(function, self.funcchooser.get(function.name))

_GLOBALPYDRIVER = PyDriver()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""conf file driver UTs."""

from unittest import main

from b3j0f.utils.ut import UTCase

from ...expr import Expression as Exp
from ..read import Read


class ReadTest(UTCase):

    def test_init_default(self):

        read = Read()

        self.assertFalse(read.select())
        self.assertFalse(read.offset())
        self.assertIsNone(read.limit())
        self.assertFalse(read.orderby())
        self.assertIsNone(read.groupby())
        self.assertFalse(read.join())

    def test_init(self):

        select = 'select'
        offset = 1
        limit = 2
        orderby = 'orderby'
        groupby = 'groupby'
        join = 'join'

        read = Read(
            select=(select,), offset=offset, limit=limit, orderby=(orderby,),
            groupby=groupby, join=join
        )

        self.assertEqual((select,), read.select())
        self.assertEqual(offset, read.offset())
        self.assertEqual(limit, read.limit())
        self.assertEqual((orderby,), read.orderby())
        self.assertEqual(groupby, read.groupby())
        self.assertEqual(join, read.join())

    def test_init_error(self):

        select = 0
        offset = '1'
        limit = '2'
        orderby = 0
        groupby = 0
        join = 0

        self.assertRaises(TypeError, Read, select=select)
        self.assertRaises(TypeError, Read, offset=offset)
        self.assertRaises(TypeError, Read, limit=limit)
        self.assertRaises(TypeError, Read, orderby=orderby)
        self.assertRaises(TypeError, Read, groupby=groupby)
        self.assertRaises(TypeError, Read, join=join)

    def test_chaining(self):

        select = 'select'
        offset = 1
        limit = 2
        orderby = 'orderby'
        groupby = 'groupby'
        join = 'join'

        read = Read()

        readbis = read.select(select).offset(offset).limit(limit)
        readbis = readbis.orderby(orderby).groupby(groupby).join(join)

        self.assertIs(read, readbis)

        self.assertEqual((select,), read.select())
        self.assertEqual(offset, read.offset())
        self.assertEqual(limit, read.limit())
        self.assertEqual((orderby,), read.orderby())
        self.assertEqual(groupby, read.groupby())
        self.assertEqual(join, read.join())

    def test___repr__(self):
        cases = [
            {
                'select': None,
                'kwargs': {},
                'where': None,
                'as': None,
                'expected': "READ ALL",
            },
            {
                'select': [Exp.r1, Exp.r2, Exp.r3],
                'kwargs': {},
                'where': None,
                'as': None,
                'expected': "READ r1, r2, r3",
            },
            {
                'select': [Exp.r],
                'kwargs': {'offset': 10},
                'where': None,
                'as': None,
                'expected': "READ r OFFSET 10",
            },
            {
                'select': [Exp.r],
                'kwargs': {'limit': 10},
                'where': None,
                'as': None,
                'expected': "READ r LIMIT 10",
            },
            {
                'select': [Exp.r],
                'kwargs': {'groupby': 'a'},
                'where': None,
                'as': None,
                'expected': "READ r GROUP BY a",
            },
            {
                'select': [Exp.r],
                'kwargs': {'orderby': [Exp.a, Exp.b]},
                'where': None,
                'as': None,
                'expected': "READ r ORDER BY a, b",
            },
            {
                'select': [Exp.r],
                'kwargs': {},
                'where': Exp.w,
                'as': None,
                'expected': "READ r WHERE w",
            },
            {
                'select': [Exp.r],
                'kwargs': {},
                'where': None,
                'as': 'a',
                'expected': "READ r AS a",
            },
        ]

        for test in cases:
            r = Read(
                select=test['select'],
                **test['kwargs']
            )

            if test['where']:
                r = r.where(test['where'])

            if test['as']:
                r = r.as_(test['as'])

            self.assertEqual(repr(r), test['expected'])

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Python utilities driver module."""

from six import iteritems

from ..request.crud.base import CRUDElement
from ..request.crud.create import Create
from ..request.crud.read import Read
from ..request.crud.update import Update
from ..request.crud.delete import Delete
from ..request.expr import Function

__all__ = ['FunctionChooser', 'SEPARATOR']


class FunctionChooser(object):

    def getfunction(self, name):

        raise NotImplementedError()


SEPARATOR = '.'  #: name separator.


def getnames(name):
    """get resource names from input name.

    :param str name: resource name from where get hierarchy of names.
    """

    return name.split(SEPARATOR)


def getchildren(elt):
    """Get children element from the request tree.

    :param BaseElement elt: element from where get children.
    :rtype: list
    """
    result = []

    if isinstance(elt, Function):

        result = elt.params

    elif isinstance(elt, CRUDElement):

        result.append(elt.query)

        if isinstance(elt, (Create, Update)):

            result.append(elt.name)

            for name, value in iteritems(elt.values):
                result.append(name)
                result.append(value)

        elif isinstance(elt, Read):

            result += elt.select()
            result += elt.orderby()
            result.append(elt.join())
            result.append(elt.limit())
            result.append(elt.offset())
            result.append(elt.groupby())

        elif isinstance(elt, Delete):

            result += elt.names()

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""request.base UTs."""

from unittest import main

from b3j0f.utils.ut import UTCase

from ..base import BaseElement


class BaseElementTest(UTCase):

    def test_init_default(self):

        base = BaseElement()

        self.assertIsNone(base.alias)

    def test_init(self):

        base = BaseElement(alias='alias')

        self.assertEqual(base.alias, 'alias')

    def test_as_(self):

        alias = BaseElement()

        alias.as_(alias='alias')

        self.assertEqual(alias.alias, 'alias')

    def test_refers(self):

        alias = BaseElement.refers('test')

        self.assertIsInstance(alias, BaseElement)
        self.assertEqual(alias.ctxname, 'test')

    def test_refers_elt(self):

        elt = BaseElement()

        alias = BaseElement.refers(elt)

        self.assertIsInstance(alias, BaseElement)
        self.assertEqual(alias.ctxname, elt.ctxname)

    def test_uuid(self):

        base0, base1 = BaseElement(), BaseElement()

        self.assertNotEqual(base0.uuid, base1.uuid)

        base0, base1 = BaseElement(uuid=1), BaseElement(uuid=1)

        self.assertEqual(base0.uuid, base1.uuid)

    def test_ctxname(self):

        class TestElement(BaseElement):
            pass

        elt = TestElement()

        self.assertEqual(elt.ctxname, elt.uuid)

        elt.alias = True

        self.assertEqual(elt.ctxname, elt.alias)

    def test_eq(self):

        class TestElement(BaseElement):
            pass

        base0, base1 = TestElement(), TestElement()

        self.assertNotEqual(base0, base1)

        base0.alias = base1.alias = 1

        self.assertEqual(base0, base1)

        base0.alias = base1.alias = None

        self.assertNotEqual(base0, base1)

        base0.uuid = base1.uuid

        self.assertEqual(base0, base1)

        base0.uuid = None

        self.assertNotEqual(base0, base1)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""conf file driver UTs."""

from unittest import main

from b3j0f.utils.ut import UTCase

from ..composite import DriverComposite, updatename

from ..py import PyDriver
from ..base import Driver
from ..ctx import Context
from ..transaction import Transaction
from ...request.expr import Expression, Function
from ...request.crud.create import Create
from ...request.crud.delete import Delete
from ...request.crud.read import Read
from ...request.crud.update import Update


class TestDriver(PyDriver):
    """Test driver."""

    def __init__(self, *args, **kwargs):

        super(TestDriver, self).__init__(*args, **kwargs)

        self.transactions = []

        self.items = [
            {'d{0}'.format(self.name): i, 'id': i} for i in range(5)
        ]

    def _process(self, transaction, *args, **kwargs):

        super(TestDriver, self)._process(
            transaction=transaction, *args, **kwargs
        )
        #print(self.name, transaction.cruds)
        transaction.ctx.setdefault('test', [self])

        self.transactions.append((transaction, kwargs))

        return transaction


class DriverCompositeTest(UTCase):
    """Tests for the driver composite."""

    def setUp(self):

        self.drivers = [
            TestDriver(
                name='d{0}'.format(i)
            ) for i in range(4)
        ]

        self.d0 = self.drivers[0]
        self.d1 = self.drivers[1]
        self.d2 = self.drivers[2]
        self.d3 = self.drivers[3]

        self.driver = DriverComposite(drivers=self.drivers)

        for driver in self.drivers:
            self.assertFalse(driver.transactions)

    def test_expr(self):

        expr = Expression.in_

        self.assertRaises(ValueError, self.driver.open(cruds=[expr]).commit)

    def test_d0(self):

        expr = Expression.d0

        self.driver.open(cruds=[expr]).commit()

        for driver in self.drivers[1:]:
            self.assertFalse(driver.transactions)

        self.assertEqual(len(self.d0.transactions), 1)

    def test_d0_dd0(self):

        expr = Expression.d0.dd0

        self.driver.open(cruds=[expr]).commit()

        for driver in self.drivers[1:]:
            self.assertFalse(driver.transactions)

        self.assertEqual(len(self.d0.transactions), 1)

    def test_expr_d0(self):

        expr = Expression.in_(Expression.d0)

        self.driver.open(cruds=[expr]).commit()

        for driver in self.drivers[1:]:
            self.assertFalse(driver.transactions)

        self.assertEqual(len(self.d0.transactions), 1)

    def test_expr_expr_d0(self):

        expr = Expression.in_(Expression.in_(Expression.d0))

        self.driver.open(cruds=[expr]).commit()

        for driver in self.drivers[1:]:
            self.assertFalse(driver.transactions)

        self.assertEqual(len(self.d0.transactions), 1)

    def test_expr_d0_expr_d0(self):

        expr = Expression.in_(Expression.d0, Expression.in_(Expression.d0))

        self.driver.open(cruds=[expr]).commit()

        for driver in self.drivers[1:]:
            self.assertFalse(driver.transactions)

        self.assertEqual(len(self.d0.transactions), 1)

    def test_expr_d0_d1_d2(self):

        expr = Expression.in_(
            Expression.d0, Expression.in_(Expression.d1), Expression.d2
        )

        self.driver.open(cruds=[expr]).commit()

        for driver in self.drivers[3:]:
            self.assertFalse(driver.transactions)

        self.assertEqual(len(self.d0.transactions), 1)
        self.assertEqual(len(self.d1.transactions), 1)
        self.assertEqual(len(self.d2.transactions), 1)

    def test_expr_d0_expr_d1_expr_d0_d1(self):

        expr = Expression.in_(
            Expression.d0, Expression.in_(
                Expression.d1, Expression.in_(
                    Expression.d0, Expression.d2
                )
            )
        )

        self.driver.open(cruds=[expr]).commit()

        for driver in self.drivers[3:]:
            self.assertFalse(driver.transactions)

        self.assertEqual(len(self.d0.transactions), 1)
        self.assertEqual(len(self.d1.transactions), 1)
        self.assertEqual(len(self.d2.transactions), 1)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""conf file driver UTs."""

from unittest import main

from b3j0f.utils.ut import UTCase

from ..base import Driver


class DriverTest(UTCase):

    def setUp(self):

        self.transactions = []
        self.lkwargs = []

        self_ = self

        class TestDriver(Driver):

            name = 'test'

            def _process(self, transaction, **kwargs):

                self_.transactions.append(transaction)
                self_.lkwargs.append(kwargs)

                return transaction

        self.drivercls = TestDriver
        self.driver = TestDriver()

        self.transaction = self.driver.open(autocommit=True)

    def test_class_name(self):

        self.assertEqual(self.drivercls.name, 'test')

    def test_custom_name(self):

        self.assertEqual(self.drivercls(name='example').name, 'example')

    def test_kwargs(self):

        kwargs = {'bar': 'foo'}

        fkwargs = kwargs.copy()
        fkwargs['async'] = False

        self.driver.process(transaction=self.transaction, **kwargs)

        self.assertEqual(self.transactions, [self.transaction])
        self.assertEqual(self.lkwargs, [fkwargs])

    def test_callback(self):

        kwargs = {'bar': 'foo'}

        fkwargs = kwargs.copy()
        fkwargs['async'] = False

        transactions = []

        def callback(transaction, **kwargs):
            transactions.append((transaction, kwargs))

        self.driver.process(
            transaction=self.transaction, callback=callback, **kwargs
        )

        self.assertEqual(self.transactions, [self.transaction])
        self.assertEqual(self.lkwargs, [fkwargs])
        self.assertEqual(transactions, [(self.transaction, fkwargs)])

    def test_async(self):

        kwargs = {'bar': 'foo'}

        fkwargs = kwargs.copy()
        fkwargs['async'] = True

        transactions = []

        def callback(transaction, **kwargs):
            transactions.append((transaction, kwargs))

        thread = self.driver.process(
            transaction=self.transaction, callback=callback, async=True,
            **kwargs
        )

        thread.join()

        self.assertEqual(self.transactions, [self.transaction])
        self.assertEqual(self.lkwargs, [fkwargs])
        self.assertEqual(transactions, [(self.transaction, fkwargs)])

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""conf file driver UTs."""

from unittest import main

from six import iteritems

from b3j0f.utils.ut import UTCase

from ..custom import (
    CreateAnnotation, ReadAnnotation, UpdateAnnotation, DeleteAnnotation,
    obj2driver, CustomDriver, query2kwargs
)
from ..ctx import Context
from ..custom import func2crudprocessing
from ..transaction import Transaction, State
from ...request.crud.base import CRUD
from ...request.crud.create import Create
from ...request.crud.delete import Delete
from ...request.crud.read import Read
from ...request.crud.update import Update
from ...request.expr import Expression


class CustomDriverTest(UTCase):

    def setUp(self):

        self.driver = CustomDriver()
        self.maxDiff = None

    def test_default(self):

        driver = CustomDriver()

        self.assertFalse(driver.creates)
        self.assertFalse(driver.reads)
        self.assertFalse(driver.updates)
        self.assertFalse(driver.deletes)
        self.assertFalse(driver.functions)

    def test_read(self):

        def read(crud, transaction, **kwargs):
            self.assertIsInstance(crud, Read)
            self.assertIs(transaction.state, State.COMMITTING)
            return [['read', transaction, kwargs]]

        driver = CustomDriver(reads=[read, read])

        transaction = driver.read()

        self.assertEqual(len(transaction.ctx), 1)

        for name, value in iteritems(transaction.ctx):
            item = ['read', transaction, {'async': False}]
            self.assertEqual(value, [item, item])

    def test_create(self):

        def create(crud, transaction, **kwargs):
            self.assertIsInstance(crud, Create)
            self.assertIs(transaction.state, State.COMMITTING)
            return [['create', transaction, kwargs]]

        driver = CustomDriver(creates=[create, create])

        transaction = driver.create(values={})

        self.assertEqual(len(transaction.ctx), 1)

        for name, value in iteritems(transaction.ctx):
            item = ['create', transaction, {'async': False}]
            self.assertEqual(value, [item, item])

    def test_update(self):

        def update(crud, transaction, **kwargs):
            self.assertIsInstance(crud, Update)
            self.assertIs(transaction.state, State.COMMITTING)
            return [['update', transaction, kwargs]]

        driver = CustomDriver(updates=[update, update])

        transaction = driver.update(values={})

        self.assertEqual(len(transaction.ctx), 1)

        for name, value in iteritems(transaction.ctx):
            item = ['update', transaction, {'async': False}]
            self.assertEqual(value, [item, item])

    def test_delete(self):

        def delete(crud, transaction, **kwargs):
            self.assertIsInstance(crud, Delete)
            self.assertIs(transaction.state, State.COMMITTING)
            return [['delete', transaction, kwargs]]

        driver = CustomDriver(deletes=[delete, delete])

        transaction = driver.delete()

        self.assertEqual(len(transaction.ctx), 1)

        for name, value in iteritems(transaction.ctx):
            item = ['delete', transaction, {'async': False}]
            self.assertEqual(value, [item, item])


class Obj2DriverTest(UTCase):

    def test_gateway(self):

        self_ = self

        class Test(object):

            def __init__(self, *args, **kwargs):

                super(Test, self).__init__(*args, **kwargs)

                self.kwargs = []

            @CreateAnnotation()
            @ReadAnnotation()
            def cr(self, a, **kwargs):

                self_.assertEqual(a, {'>': [2]})

                return [{'cr': kwargs}]

            @ReadAnnotation()
            @UpdateAnnotation()
            def ru(self, a, **kwargs):

                self_.assertEqual(a, {'>': [2]})

                return [{'ru': kwargs}]

            @UpdateAnnotation()
            @DeleteAnnotation()
            def ud(self, a, **kwargs):

                self_.assertEqual(a, {'>': [2]})

                return [{'ud': kwargs}]

            @DeleteAnnotation()
            @CreateAnnotation()
            def dc(self, a, **kwargs):

                self_.assertEqual(a, {'>': [2]})

                return [{'dc': kwargs}]

            def test(self, **kwargs):

                self.kwargs.append(kwargs)

                return 'test'

        test = Test()

        driver = obj2driver(test)

        query = (Expression.a > 2) & Expression.test(Expression.test())

        transaction = driver.create(values={}, query=query)

        self.assertEqual(len(transaction.ctx), 3)
        self.assertEqual(len(test.kwargs), 2)

        transaction = driver.read(query=query)

        self.assertEqual(len(transaction.ctx), 6)
        self.assertEqual(len(test.kwargs), 4)

        transaction = driver.update(values={}, query=query)

        self.assertEqual(len(transaction.ctx), 3)
        self.assertEqual(len(test.kwargs), 6)

        transaction = driver.delete(query=query)

        self.assertEqual(len(transaction.ctx), 3)
        self.assertEqual(len(test.kwargs), 8)

    def test_gateway_or(self):

        self_ = self

        class Test(object):

            def __init__(self, *args, **kwargs):

                super(Test, self).__init__(*args, **kwargs)

                self.kwargs = []

            @CreateAnnotation()
            @ReadAnnotation()
            def cr(self, a=None, b=None, **kwargs):

                if a is not None:
                    self_.assertEqual(a, {'>': [2]})

                if b is not None:
                    self_.assertEqual(b, {'>': [3]})

                return [{'cr': kwargs}]

            @ReadAnnotation()
            @UpdateAnnotation()
            def ru(self, a=None, b=None, **kwargs):

                if a is not None:
                    self_.assertEqual(a, {'>': [2]})

                if b is not None:
                    self_.assertEqual(b, {'>': [3]})

                return [{'ru': kwargs}]

            @UpdateAnnotation()
            @DeleteAnnotation()
            def ud(self, a=None, b=None, **kwargs):

                if a is not None:
                    self_.assertEqual(a, {'>': [2]})

                if b is not None:
                    self_.assertEqual(b, {'>': [3]})

                return [{'ud': kwargs}]

            @DeleteAnnotation()
            @CreateAnnotation()
            def dc(self, a=None, b=None, **kwargs):

                if a is not None:
                    self_.assertEqual(a, {'>': [2]})

                if b is not None:
                    self_.assertEqual(b, {'>': [3]})

                return [{'dc': kwargs}]

            def test(self, **kwargs):

                self.kwargs.append(kwargs)

                return 'test'

        test = Test()

        driver = obj2driver(test)

        query = (Expression.a > 2) & Expression.test() | (Expression.b > 3)

        transaction = driver.create(values={}, query=query)

        self.assertEqual(len(transaction.ctx), 2)
        self.assertEqual(len(test.kwargs), 1)

        transaction = driver.read(query=query)

        self.assertEqual(len(transaction.ctx), 3)
        self.assertEqual(len(test.kwargs), 2)

        transaction = driver.update(values={}, query=query)

        self.assertEqual(len(transaction.ctx), 2)
        self.assertEqual(len(test.kwargs), 3)

        transaction = driver.delete(query=query)

        self.assertEqual(len(transaction.ctx), 2)
        self.assertEqual(len(test.kwargs), 4)

    def test_notgateway(self):

        class Test(object):

            def __init__(self, *args, **kwargs):

                super(Test, self).__init__(*args, **kwargs)

                self.kwargs = []

            @CreateAnnotation(gateway=False)
            @ReadAnnotation(gateway=False)
            def cr(self, **kwargs):

                return

            @ReadAnnotation(gateway=False)
            @UpdateAnnotation(gateway=False)
            def ru(self, **kwargs):

                return

            @UpdateAnnotation(gateway=False)
            @DeleteAnnotation(gateway=False)
            def ud(self, **kwargs):

                return

            @DeleteAnnotation(gateway=False)
            @CreateAnnotation(gateway=False)
            def dc(self, **kwargs):

                return

            def test(self, **kwargs):

                self.kwargs.append(kwargs)

                return 'test'

        test = Test()

        driver = obj2driver(test)

        query = Expression.a > Expression.test(Expression.test())

        transaction = driver.create(values={}, query=query)

        self.assertEqual(len(transaction.ctx), 3)
        self.assertEqual(len(test.kwargs), 2)

        transaction = driver.read(query=query)

        self.assertEqual(len(transaction.ctx), 3)
        self.assertEqual(len(test.kwargs), 4)

        transaction = driver.update(values={}, query=query)

        self.assertEqual(len(transaction.ctx), 3)
        self.assertEqual(len(test.kwargs), 6)

        transaction = driver.delete(query=query)

        self.assertEqual(len(transaction.ctx), 3)
        self.assertEqual(len(test.kwargs), 8)


class Query2KwargsTest(UTCase):

    def test_default(self):

        allkwargs = query2kwargs(query=None, ctx=Context(), pnames=[])

        self.assertEqual(allkwargs, [{}])

    def test_expr(self):

        allkwargs = query2kwargs(query=Expression.a, ctx=Context(), pnames=[])

        self.assertEqual(allkwargs, [{}])

    def test_expr_a(self):

        allkwargs = query2kwargs(
            query=Expression.a, ctx=Context(), pnames=['a']
        )

        self.assertEqual(allkwargs, [{'a': {'exists': []}}])

    def test_expr_a_b(self):

        allkwargs = query2kwargs(
            query=Expression.a & Expression.b & Expression.c,
            ctx=Context(), pnames=['a', 'b']
        )

        self.assertEqual(
            allkwargs,
            [{'a': {'exists': []}, 'b': {'exists': []}}]
        )

    def test_and(self):

        allkwargs = query2kwargs(
            query=(Expression.a > 2) & (Expression.a < 3),
            ctx=Context(),
            pnames=['a']
        )

        self.assertEqual(
            allkwargs,
            [{'a': {'>': [2], '<': [3]}}]
        )

    def test_or(self):

        allkwargs = query2kwargs(
            query=(Expression.a > 2) | (Expression.a < 3),
            ctx=Context(),
            pnames=['a']
        )

        self.assertEqual(
            allkwargs,
            [{'a': {'>': [2]}}, {'a': {'<': [3]}}]
        )

    def test_or_and(self):

        query = (Expression.a > 2) | ((Expression.a < 3) & (Expression.a == 4))

        allkwargs = query2kwargs(
            query=query,
            ctx=Context(),
            pnames=['a']
        )

        self.assertEqual(
            allkwargs,
            [{'a': {'>': [2]}}, {'a': {'<': [3], '==': [4]}}]
        )

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Module which specifices a composite of drivers."""

from inspect import getmembers

from b3j0f.schema import Schema, data2schema

from six import iteritems

from copy import deepcopy

from .base import Driver
from .utils import getnames, getchildren
from .ctx import Context
from ..request.consts import FuncName
from ..request.base import BaseElement
from ..request.crud.base import CRUDElement
from ..request.crud.read import Read
from ..request.expr import Expression, Function

__all__ = ['DriverComposite']


class DriverComposite(Driver):
    """In charge of distributing a request to several drivers.

    Driver parameters are :

    .. csv-table::

        :header: name, type, description

        - discovery, bool (False), enable driver discovery from queries.
        - many, bool (False), disable multi driver execution for one query.
        - besteffort, bool (False), enable discovery and many.
        - maxdepth, int (3), max depth data search in the driver tree.
        - explain, bool (False), return query information processsing in ctx.
    """

    def __init__(self, drivers, *args, **kwargs):
        """
        :param list drivers: drivers to use.
        """
        super(DriverComposite, self).__init__(*args, **kwargs)

        self.drivers = {}
        self.sdrivers = {}

        for driver in drivers:
            if isinstance(driver, Schema):
                sdriver = driver

            else:
                sdriver = data2schema(driver, name=driver.name, _force=True)

            self.drivers[driver.name] = driver
            self.sdrivers[driver.name] = sdriver

    def getdrivers(self, name, maxdepth=3, discovery=False, many=False):
        """Get a list of drivers corresponding with input model name.

        :param str name: data name to identify such as a driver model.
        :param bool discovery: if True (False by default), try to find drivers
            where name match with driver models.
        :param bool many: if False (default), raise a ValueError if more than
            one driver is found.
        :return: list of couples of driver with model where name match a
            driver/model name.
        :rtype: list
        :raises: ValueError if name is not a standard query name and no driver
            is found.
        """
        result = []

        if name in self.supportedfunctions:
            return result

        names = getnames(name)

        rootname = names[0]

        if discovery:

            tmpelts = []  # list of couple of driver/model.
            elts = []  # list of couple of driver/model.

            for depth in range(maxdepth + 1):

                if elts:
                    tmpelts = []

                    for driver, model in list(elts):
                        if hasattr(model, rootname):
                            tmpelts.append((driver, getattr(model, rootname)))

                        elif isinstance(model, Schema):
                            for mname, submodel in iteritems(
                                model.getschemas()
                            ):
                                tmpelts.append((driver, submodel))

                        else:
                            for name, member in getmembers(model):
                                if name[0] != '_':
                                    tmpelts.append((driver, member))

                    elts = tmpelts

                else:
                    if rootname in self.drivers:
                        result = [self.drivers[rootname]]
                        break

                    else:
                        elts = [
                            (self.drivers[name], self.sdrivers[name])
                            for name in self.drivers
                        ]

            if not result:
                if elts:
                    for name in names[1:]:
                        elts = [
                            (elt[0], getattr(elt[1], name)) for elt in elts
                            if hasattr(elt[1], name)
                        ]
                        if not elts:
                            break

                    else:
                        result = [item[0] for item in elts]

        else:
            if rootname in self.drivers:
                result = [self.drivers[rootname]]

            else:
                raise ValueError(
                    '{0} is not handled by {1}'.format(name, self)
                )

        if result:
            if many and len(result) > 1:
                raise ValueError(
                    'Too many drivers found for elt {0}: {1}.'.format(
                        elt, result
                    )
                )

        else:
            raise ValueError('No driver found for processing {0}'.format(name))

        return result

    def _process(self, transaction, **kwargs):

        for crud in transaction.cruds:
            self.processdeeply(elt=crud, transaction=transaction, **kwargs)

        return transaction

    def processdeeply(self, elt, transaction, _elts=None, **kwargs):
        """Parse input elt and return its evaluation."""
        result = elt

        # process specific driver parameters
        discovery = kwargs.get('discovery', False)
        many = kwargs.get('many', False)
        besteffort = kwargs.get('besteffort')
        maxdepth = kwargs.get('maxdepth', 3)

        if besteffort is not None:
            many = discovery = besteffort

        ctx = transaction.ctx

        if elt in ctx:
            result = ctx[elt]

        elif isinstance(elt, BaseElement):
            # get driver and model

            # default case, elt is a crudelement or elt is a standard query
            if isinstance(elt, CRUDElement):
                drivers = []

            else:
                drivers = self.getdrivers(
                    name=elt.name, maxdepth=maxdepth, discovery=discovery,
                    many=many
                )

            # fill elts
            if _elts is None:
                _elts = [[drivers, elt]]

            elif drivers:
                olddrivers = _elts[-1][0]
                if not olddrivers:
                    _elts[-1][0] = drivers

                elif olddrivers != drivers:
                        _elts.append([drivers, elt])

            children = getchildren(elt)

            isor = isinstance(elt, Function) and elt.name == FuncName.OR.value

            for child in children:

                if isor:
                    ctx = Context(transaction.ctx)
                    ftransaction = transaction.open(ctx=ctx)

                else:
                    ftransaction = transaction

                self.processdeeply(
                    elt=child, transaction=ftransaction, _elts=_elts, **kwargs
                )

                if isor:
                    transaction.ctx.fill(ftransaction.ctx)

            if _elts[-1][1].ctxname == elt.ctxname:
                drivers, _ = _elts.pop()

                if drivers:

                    if isinstance(elt, Expression):
                        crud = Read(query=elt)

                    else:
                        crud = elt

                    crudcopy = crud.copy()

                    threads = []

                    for driver in drivers:
                        updatename(elt=crudcopy, driver=driver)

                        dparams = deepcopy(kwargs)
                        dparams.pop('async', None)

                        transaction.ctx[elt] = []

                        ftransaction = transaction.open(
                            driver=driver, cruds=[crudcopy]
                        )

                        def callback(transaction, **kwargs):
                            transaction.ctx[elt] += transaction.ctx[crudcopy]

                        if len(drivers) == 1:
                            result = ftransaction.commit(
                                callback=callback, **dparams
                            )

                        else:
                            thread = ftransaction.commit(
                                async=True, callback=callback, **dparams
                            )
                            threads.append(thread)

                        for thread in threads:
                            thread.join()

                    result = ctx[elt]

                else:
                    raise ValueError(
                        'No driver found to process {0}'.format(elt)
                    )

        return result

    def __repr__(self):
        """Driver representation with drivers."""
        return 'CompositeDriver({0}, {1})'.format(self.name, self.drivers)


def updatename(elt, driver):
    """Rename elt in order to be specific to input driver."""
    for slot in elt.__slots__:

        subelt = getattr(elt, slot)

        if isinstance(subelt, Expression):
            if subelt.name.startswith(driver.name):
                subelt.name = subelt.name[len(driver.name) + 1:]

        elif isinstance(subelt, CRUDElement):

            subelt = updatename(elt=subelt, driver=driver)
            setattr(elt, slot, subelt)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Base crud module."""

from enum import IntEnum, unique

from ..expr import BaseElement

__all__ = ['CRUDElement', 'CRUD']


@unique
class CRUD(IntEnum):

    CREATE = 1
    READ = 2
    UPDATE = 3
    DELETE = 4


class CRUDElement(BaseElement):
    """Base crud operation.

    Can be associated to a transaction."""

    __slots__ = ['transaction', 'query', 'dparams'] + BaseElement.__slots__

    def __init__(
            self, query=None, transaction=None, dparams=None, *args, **kwargs
    ):
        """
        :param Expression query: query.
        :param Transaction transaction: related transaction.
        :param dict dparams: driver params.
        """

        super(CRUDElement, self).__init__(*args, **kwargs)

        self.query = query
        self.transaction = transaction
        self.dparams = dparams

    def __call__(self, **kwargs):
        """Execute this CRUD element.

        :param dict kwargs: driver specific kwargs. See Driver.process for more
            details.
        :return: this execution result if async is False.
        """
        if self.transaction is None:
            raise RuntimeError('No transaction attached.')

        else:
            self.transaction.process(cruds=[self], **kwargs)

            return self.transaction.ctx.get(self)

    def where(self, *query):
        """Getter/Setter for query.

        Setter if query is given. Getter otherwise.

        :return: self if query is given, self query otherwise.
        :rtype: Request or BaseElement"""

        if query:
            query = query[0]

            if not isinstance(query, BaseElement):
                raise TypeError('{0}. {1} expected'.format(query, BaseElement))

            if self.query is None:
                self.query = query

            else:
                self.query &= query

            result = self

        else:
            result = self.query

        return result

    def orwhere(self, query):
        """Apply input query to this query.

        :return: self
        :rtype: Request
        """
        if self.query is None:
            self.query = query

        else:
            self.query |= query

        return self

    def __iand__(self, other):

        return self.where(other)

    def __ior__(self, other):

        return self.orwhere(other)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Transaction module."""

from uuid import uuid4 as uuid

from enum import IntEnum, unique

from .ctx import Context

from ..request.crud.create import Create
from ..request.crud.delete import Delete
from ..request.crud.read import Read
from ..request.crud.update import Update

from ..utils import tostr

__all__ = ['Transaction', 'State']


@unique
class State(IntEnum):
    """Transaction state."""

    PENDING = 1
    COMMITTING = 2
    ROLLBACKING = 3


class Transaction(object):
    """Request execution context."""

    __slots__ = [
        'driver', 'parent', 'cruds', 'autocommit', 'uuid', 'state', 'ctx',
        'dparams'
    ]

    def __init__(
            self, driver,
            parent=None, autocommit=False, ctx=None, cruds=None, dparams=None,
            *args, **kwargs
    ):
        """
        :param Driver driver: data base driver.
        :param Transaction parent: parent transaction.
        :param bool autocommit: if True (default False), commit as soon a CRUD
            operation is processed by this transaction.
        :param Context ctx: CRUD execution context.
        :param list cruds: crud operations.
        :param dict dparams: driver specific parameters.
        """

        super(Transaction, self).__init__(*args, **kwargs)

        self.driver = driver
        self.uuid = uuid()
        self.parent = parent
        self.cruds = [] if cruds is None else cruds
        self.state = State.COMMITTING if autocommit else State.PENDING
        self.autocommit = autocommit
        self.ctx = Context() if ctx is None else ctx
        self.dparams = dparams

    def __enter__(self):

        return self

    def __exit__(self, exc_type, exc_value, exc_tb):

        if exc_type is None:
            self.commit()

        else:
            self.rollback()
            raise

    def __repr__(self):

        return tostr(self)

    def commit(self, **kwargs):
        """Commit this transaction and return this.

        While committing, this transaction change of state. After that, this
        transaction state go back to PENDING.
        :return: driver processing result."""

        self.state = State.COMMITTING

        result = self.driver.process(transaction=self, **kwargs)

        if not self.autocommit:
            self.state = State.PENDING

        return result

    def rollback(self, **kwargs):
        """Rollback this transaction and return this.

        While rollbacking, this transaction change of state. After that, this
        transaction state go back to PENDING.

        :param dict kwargs: driver kwargs.
        :return: driver processing result."""

        self.state = State.ROLLBACKING

        result = self.driver.process(transaction=self, **kwargs)

        self.state = State.COMMITTING if self.autocommit else State.PENDING

        return result

    def process(self, cruds=None, **kwargs):
        """Process input cruds with control paremeters.

        :param dict kwargs: driver specific kwargs. See Driver.process for more
            details.
        :return: process execution."""

        if cruds is not None:
            self.cruds += cruds

        if self.autocommit:
            self.state = State.COMMITTING

        result = self.driver.process(transaction=self, **kwargs)

        if self.parent is not None:
            if self.state is State.COMMITTING:
                result = self.parent.commit()

            elif self.state is State.ROLLBACKING:
                result = self.parent.rollback()

        return result

    def open(self, autocommit=None, cruds=None, driver=None, ctx=None):

        if autocommit is None:
            autocommit = self.autocommit

        if driver is None:
            driver = self.driver

        if ctx is None:
            ctx = self.ctx

        return Transaction(
            driver=driver, parent=self, ctx=ctx,
            autocommit=autocommit, cruds=cruds
        )

    def _process(self, cls, **kwargs):
        """Custom processing."""
        crud = cls(transaction=self, **kwargs)

        self.process(cruds=[crud])

        return self

    def create(self, **kwargs):
        """Quick creation."""
        return self._process(cls=Create, **kwargs)

    def read(self, **kwargs):
        """Quick reading."""
        return self._process(cls=Read, **kwargs)

    def update(self, **kwargs):
        """Quick updating."""
        return self._process(cls=Update, **kwargs)

    def delete(self, **kwargs):
        """Quick deletion."""
        return self._process(cls=Delete, **kwargs)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Read module."""

from collections import Iterable

from six import string_types, iteritems

from .base import CRUDElement

from .join import Join

__all__ = ['Read', 'Cursor']


class Read(CRUDElement):
    """In charge of parameterize a reading request.

    Execution is done in calling it or in using the getslice method.
    Result is a Cursor."""

    __slots__ = [
        '_select', '_offset', '_limit', '_orderby', '_groupby', '_join'
    ] + CRUDElement.__slots__

    def __init__(
            self,
            select=None, offset=None, limit=None, orderby=None, groupby=None,
            join=None, *args, **kwargs
    ):
        """
        :param list select: data to select.
        :param int offset: data to avoid.
        :param int limit: max number of data to retrieve.
        :param str orderby: data sorting.
        :param list groupby: data field group.
        :param join: join type (INNER, LEFT, etc.).
        :type join: str or Join
        """

        super(Read, self).__init__(*args, **kwargs)

        # initialize protected properties
        self._select = ()
        self._offset = None
        self._limit = None
        self._orderby = ()
        self._groupby = None
        self._join = None

        # set parameters
        if select is not None:
            self.select(*select)

        if offset is not None:
            self.offset(offset)

        if limit is not None:
            self.limit(limit)

        if orderby is not None:
            self.orderby(*orderby)

        if groupby is not None:
            self.groupby(groupby)

        if join is not None:
            self.join(join)

    def offset(self, *value):
        """Get or set offset if value is not None.

        :param int value: value to set. Default is None.
        :return: depending on value. If None, return this offset, otherwise
            this.
        :rtype: int or Read
        """
        if value:
            value = value[0]

            if not isinstance(value, int):
                raise TypeError(
                    'Wrong value {0}. {1} expected'.format(value, int)
                )

            self._offset = value
            result = self

        else:
            result = self._offset

        return result

    def limit(self, *value):
        """Get or set limit if value is not None.

        :param int value: value to set. Default is None.
        :return: depending on value. If None, return this offset, otherwise
            this.
        :rtype: int or Read
        """
        if value:
            value = value[0]

            if not isinstance(value, int):
                raise TypeError(
                    'Wrong value {0}. {1} expected'.format(value, int)
                )

            result = self
            self._limit = value

        else:
            result = self._limit

        return result

    def orderby(self, *values):
        """Get or set orderby if value is not None.

        :param tuple value: value to set. Default is None.
        :return: depending on value. If None, return this offset, otherwise
            this.
        :rtype: tuple or Read
        """
        if values:
            self._orderby = values
            result = self

        else:
            result = self._orderby

        return result

    def groupby(self, *value):
        """Get or set groupby if value exists.

        :param int value: value to set. Default is None.
        :return: depending on value. If None, return this offset, otherwise
            this.
        :rtype: int or Read
        """
        if value:
            value = value[0]

            if not isinstance(value, string_types):
                raise TypeError(
                    'Wrong value {0}. {1} expected'.format(value, str)
                )

            result = self
            self._groupby = value

        else:
            result = self._groupby

        return result

    def select(self, *values):
        """Get or set select if value is not None.

        :param tuple value: value to set. Default is None.
        :return: depending on value. If None, return this offset, otherwise
            this.
        :rtype: tuple or Read
        """
        if values:
            self._select = values
            result = self

        else:
            result = self._select

        return result

    def join(self, *value):
        """Get or set join if value is not None.

        :param value: value to set. Default is None.
        :type value: str or Join
        :return: depending on value. If None, return this offset, otherwise
            this.
        :rtype: str or Join or Read
        """
        if value:
            value = value[0]

            if not isinstance(value, string_types + (Join,)):
                raise TypeError(
                    'Wrong value {0}. {1} expected'.format(
                        value, string_types + (Join,)
                    )
                )

            self._join = value.name if isinstance(value, Join) else value
            result = self

        else:
            result = self._join

        return result

    def __getslice__(self, start, stop):
        """Set offset and limit and execute the selection.

        :param int start: offset property.
        :param int stop: limit property.
        :return: selection execution result.
        :rtype: Cursor"""

        if start is not None:
            self._offset = start

        if stop is not None:
            self._limit = stop

        return self()

    def __getitem__(self, key):

        if not isinstance(key, slice):
            key = slice(key, key + 1)

        return self.__getslice__(key.start, key.stop)

    def __repr__(self):

        if self._select:
            items = [repr(item) for item in self._select]
            select = ', '.join(items)

        else:
            select = 'ALL'

        result = 'READ {0} '.format(select)

        if self._limit or self._offset or self._groupby or self._orderby:

            if self._limit is not None:
                result += 'LIMIT {0} '.format(repr(self._limit))

            if self._offset is not None:
                result += 'OFFSET {0} '.format(repr(self._offset))

            if self._groupby:
                result += 'GROUP BY {0} '.format(self._groupby)

            if self._orderby:
                items = [repr(item) for item in self._orderby]
                result += 'ORDER BY {0} '.format(', '.join(items))

        if self.query:
            result += 'WHERE {0} '.format(repr(self.query))

        if self.dparams:
            result += 'WITH '

            dparams = []
            for name, value in iteritems(self.dparams):
                dparam = '{0}'.format(name)

                if value is not True:
                    dparam += ': {0}'.format(value)

                dparams.append(dparam)

            result += '{0} '.format(', '.join(dparams))

        if self.alias:
            result += 'AS {0}'.format(self.alias)

        if result[-1] == ' ':
            result = result[:-1]

        return result


class Cursor(Iterable):
    """Read request result."""

    def __init__(self, cursor, *args, **kwargs):

        super(Cursor, self).__init__(*args, **kwargs)

        self._cursor = cursor

    def __len__(self):

        return len(self._cursor)

    def __iter__(self):

        return iter(self._cursor)

    def __getitem__(self, key):

        return self._cursor[key]

    def __getslice__(self, i, j):

        return self._cursor[i:j]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""transaction.base UTs."""

from unittest import main

from b3j0f.utils.ut import UTCase

from ..base import Driver

from ..transaction import State, Transaction

from ...request.crud.base import CRUDElement
from ...request.crud.create import Create
from ...request.crud.delete import Delete
from ...request.crud.read import Read
from ...request.crud.update import Update


class TransactionTest(UTCase):

    def setUp(self):

        self.transactions = []
        self.states = []

        self_ = self

        class TestDriver(Driver):

            def process(self, transaction, **kwargs):

                self_.states.append(transaction.state)
                self_.transactions.append(transaction)
                for crud in transaction.cruds:
                    self_.transaction.ctx[crud] = crud

        self.driver = TestDriver()

        self.transaction = Transaction(driver=self.driver)

    def test_init_default(self):

        transaction = Transaction(driver=None)

        self.assertIsNone(transaction.driver)
        self.assertEqual(transaction.ctx, {})

    def test_init_errorquery(self):

        self.assertRaises(TypeError, Transaction, query=1)

    def test_init_errorcruds(self):

        self.assertRaises(TypeError, Transaction, cruds=[1])

    def test_processcrud(self):

        crud = CRUDElement()

        self.transaction.process(cruds=[crud])

        self.assertIs(self.transaction.state, State.PENDING)
        self.assertIn(self.transaction, self.transactions)
        self.assertIn(crud, self.transaction.ctx)

    def test_create(self):

        name = 'test'
        values = {'a': 1, 'b': 2}

        self.transaction.create(name=name, values=values)

        self.assertIn(self.transaction, self.transactions)
        for key in self.transaction.ctx:
            crud = self.transaction.ctx[key]
            break

        self.assertIsInstance(crud, Create)
        self.assertIs(crud.transaction, self.transaction)
        self.assertEqual(crud.name, name)
        self.assertEqual(crud.values, values)

    def test_read(self):

        select = ('test',)
        limit = 1

        self.transaction.read(select=select, limit=limit)

        self.assertIn(self.transaction, self.transactions)

        for key in self.transaction.ctx:
            crud = self.transaction.ctx[key]
            break

        self.assertIsInstance(crud, Read)
        self.assertIs(crud.transaction, self.transaction)
        self.assertEqual(crud.select(), select)
        self.assertEqual(crud.limit(), limit)

    def test_update(self):

        name = 'test'
        values = {'a': 1, 'b': 2}

        self.transaction.update(name=name, values=values)

        self.assertIn(self.transaction, self.transactions)

        for key in self.transaction.ctx:
            crud = self.transaction.ctx[key]
            break

        self.assertIsInstance(crud, Update)
        self.assertIs(crud.transaction, self.transaction)
        self.assertEqual(crud.name, name)
        self.assertEqual(crud.values, values)

    def test_delete(self):

        names = ('test',)

        self.transaction.delete(names=names)

        self.assertIn(self.transaction, self.transactions)

        for key in self.transaction.ctx:
            crud = self.transaction.ctx[key]
            break

        self.assertIsInstance(crud, Delete)
        self.assertIs(crud.transaction, self.transaction)
        self.assertEqual(crud.names(), names)

    def test_open(self):

        newtransaction = self.transaction.open()

        self.assertIsInstance(newtransaction, Transaction)
        self.assertIsNot(newtransaction, self.transaction)
        self.assertIs(newtransaction.parent, self.transaction)
        self.assertIs(newtransaction.ctx, self.transaction.ctx)

    def test_rollback(self):

        self.assertIs(self.transaction.state, State.PENDING)

        self.assertEqual(self.states, [])

        self.transaction.process(cruds=[])

        self.assertEqual(self.states, [State.PENDING])

        self.assertIs(self.transaction.state, State.PENDING)

        self.transaction.commit()

        self.assertEqual(self.states[-1], State.COMMITTING)

        self.assertIs(self.transaction.state, State.PENDING)

        self.transaction.rollback()

        self.assertEqual(self.states[-1], State.ROLLBACKING)

        self.assertIs(self.transaction.state, State.PENDING)

    def test_enter_exit(self):

        self.assertFalse(self.states)

        with self.transaction:
            pass

        self.assertEqual(self.states[-1], State.COMMITTING)

        self.assertIs(self.transaction.state, State.PENDING)

        try:
            with self.transaction:
                raise RuntimeError()

        except:
            pass

        self.assertEqual(self.states[-1], State.ROLLBACKING)

        self.assertIs(self.transaction.state, State.PENDING)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Macro module."""

from enum import Enum, unique

__all__ = ['FuncName', 'CONDITIONS']


@unique
class FuncName(Enum):
    """Default function names which might be supported by drivers."""

    AND = '&'
    IAND = '&='
    OR = '|'
    IOR = '|='
    XOR = '^'
    IXOR = '^='
    TRUTH = 'truth'
    IS = 'is'
    ISNOT = 'isnot'
    NOT = '!'
    EQ = '=='
    NE = '!='
    GT = '>'
    GE = '>='
    LT = '<'
    LE = '<='
    ADD = '+'
    IADD = '+='
    SUB = '-'
    ISUB = '-='
    DIV = '/'
    IDIV = '/='
    FLOORDIV = '//'
    IFLOORDIV = '//='
    MUL = '*'
    IMUL = '*='
    POW = '**'
    IPOW = '**='
    LIKE = '%%'
    RSHIFT = '>>'
    IRSHIFT = '>>='
    LSHIFT = '<<'
    ILSHIFT = '<<='
    MOD = '%'
    IMOD = '%='
    NEG = 'neg'
    ABS = 'abs'
    INVERT = '~'
    EXISTS = 'exists'
    NEXISTS = 'nexists'
    GETSLICE = 'getslice'
    SETSLICE = 'setslice'
    DELSLICE = 'delslice'
    GETITEM = 'getitem'
    SETITEM = 'setitem'
    DELITEM = 'delitem'

    # remainders functions are not supported by the Expression methods
    ISNULL = 'isnull'
    BETWEEN = 'between'
    IN = 'in'
    NIN = 'nin'

    # selection operators  TODO: might be migrated to the Read object...
    HAVING = 'having'
    UNION = 'union'
    INTERSECT = 'intersect'

    # request comparison
    ALL = 'all'
    ANY = 'any'
    SOME = 'some'

    # DB operations
    OPTIMIZE = 'optimize'
    VERSION = 'version'

    # aggregation operations
    AVG = 'avg'
    COUNT = 'count'
    MEAN = 'mean'
    MAX = 'max'
    MIN = 'min'
    SUM = 'sum'

    # string operations
    CONCAT = 'concat'
    ICONCAT = 'iconcat'
    LENGTH = 'length'
    REPLACE = 'replace'
    SOUNDEX = 'soundex'
    SUBSTR = 'substr'
    SUBSTRING = 'substring'
    LEFT = 'left'
    RIGHT = 'right'
    REVERSE = 'reverse'
    TRIM = 'trim'
    LTRIM = 'ltrim'
    RTRIM = 'rtrim'
    LPAD = 'lpad'
    RPAD = 'rpad'
    UPPER = 'upper'
    LOWER = 'lower'
    UCASE = 'ucase'
    LCASE = 'lcase'
    LOCATE = 'locate'
    INSTR = 'instr'

    # mathematical operations
    RAND = 'rand'
    ROUND = 'round'
    MD5 = 'md5'

    # datetime operations
    NOW = 'now'
    SEC_TO_TIME = 'sec_to_time'
    DATEDIFF = 'datediff'
    MONTH = 'month'
    YEAR = 'year'

    # array operations
    INDEX = 'index'
    REPEAT = 'repeat'
    IREPEAT = 'irepeat'
    COUNTOF = 'countof'
    INCLUDE = 'include'

    # additional operations
    CAST = 'cast'
    CONVERT = 'convert'
    GROUPCONCAT = 'groupconcat'

    @staticmethod
    def contains(value):
        """True iif input value is a value of FuncName.

        :param str value: funcname value to check.
        :rtype: bool
        """
        result = False

        for member in FuncName.__members__.values():
            if member.value == value:
                result = True
                break

        return result


CONDITIONS = [  #: expression conditions.

    FuncName.OR.value,
    FuncName.AND.value,

    FuncName.IS.value,
    FuncName.ISNOT.value,
    FuncName.NOT.value,
    FuncName.EQ.value,
    FuncName.NE.value,

    # numerical functions
    FuncName.GT.value,
    FuncName.GE.value,
    FuncName.LT.value,
    FuncName.LE.value,

    FuncName.LIKE.value,

    FuncName.EXISTS.value,
    FuncName.NEXISTS.value,

    # remainders functions are not supported by the Expression methods
    FuncName.ISNULL.value,
    FuncName.BETWEEN.value,
    FuncName.IN.value,

    # selection operators  TODO: might be migrated to the Read object...
    FuncName.HAVING.value,

    # request comparison
    FuncName.ALL.value,
    FuncName.ANY.value,
    FuncName.SOME.value
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""conf file driver UTs."""

from unittest import main

from b3j0f.utils.ut import UTCase

from ..consts import FuncName
from ..expr import Expression, Function


class ExpressionTest(UTCase):

    def test_init(self):

        name = 'test'

        expr = Expression(name=name)

        self.assertEqual(name, expr.name)
        self.assertIsNone(expr.alias)

    def test_getattr_cls_reserved(self):

        expr = Expression.is_

        self.assertEqual(expr.name, 'is')

    def test_getattr_cls_attr(self):

        expr = Expression.name_

        self.assertEqual(expr.name, 'name')

    def test_getattr(self):

        expr = Expression.A.B.C

        self.assertEqual(expr.name, 'A.B.C')

    def test_getattr_attr(self):

        expr = Expression.A.name_

        self.assertEqual(expr.name, 'A.name')

    def test_getattr_reserved(self):

        expr = Expression.A.is_

        self.assertEqual(expr.name, 'A.is')

    def _assertfunc(self, func, funcname, *params):

        self.assertIsInstance(func, Function)
        self.assertEqual(funcname.value, func.name)
        self.assertEqual(list(params), func.params)

    def _asserttype(self, expr, method, other=None):

        self.assertRaises(TypeError, getattr(expr, method), other)

    def test__and__(self):

        expr0, expr1 = Expression(name=''), Expression(name='')

        func = expr0 & expr1

        self._assertfunc(func, FuncName.AND, expr0, expr1)

    def test__rand__(self):

        expr0, expr1 = True, Expression(name='')

        func = expr0 & expr1

        self._assertfunc(func, FuncName.AND, expr1, expr0)

    def test__or__(self):

        expr0, expr1 = Expression(name=''), Expression(name='')

        func = expr0 | expr1

        self._assertfunc(func, FuncName.OR, expr0, expr1)

    def test__ror__(self):

        expr0, expr1 = True, Expression(name='')

        func = expr0 | expr1

        self._assertfunc(func, FuncName.OR, expr1, expr0)

    def test__xor__(self):

        expr0, expr1 = Expression(name=''), Expression(name='')

        func = expr0 ^ expr1

        self._assertfunc(func, FuncName.XOR, expr0, expr1)

    def test__rxor__(self):

        expr0, expr1 = True, Expression(name='')

        func = expr0 ^ expr1

        self._assertfunc(func, FuncName.XOR, expr1, expr0)

    def test__gt__(self):

        expr = Expression(name='')

        func = expr > 1

        self._assertfunc(func, FuncName.GT, expr, 1)

        self._asserttype(expr, '__gt__')

    def test__ge__(self):

        expr = Expression(name='')

        func = expr >= 1

        self._assertfunc(func, FuncName.GE, expr, 1)
        self._asserttype(expr, '__ge__')

    def test__lt__(self):

        expr = Expression(name='')

        func = expr < 1

        self._assertfunc(func, FuncName.LT, expr, 1)
        self._asserttype(expr, '__lt__')

    def test__le__(self):

        expr = Expression(name='')

        func = expr <= 1

        self._assertfunc(func, FuncName.LE, expr, 1)
        self._asserttype(expr, '__le__')

    def test__eq__(self):

        expr0, expr1 = Expression(name=''), Expression(name='')

        func = expr0 == expr1

        self._assertfunc(func, FuncName.EQ, expr0, expr1)

    def test__ne__(self):

        expr0, expr1 = Expression(name=''), Expression(name='')

        func = expr0 != expr1

        self._assertfunc(func, FuncName.NE, expr0, expr1)

    def test__add__(self):

        expr = Expression(name='')

        func = expr + 1

        self._assertfunc(func, FuncName.ADD, expr, 1)
        self._asserttype(expr, '__add__')

    def test_concat(self):

        expr = Expression(name='')

        func = expr + ''

        self._assertfunc(func, FuncName.CONCAT, expr, '')
        self._asserttype(expr, '__add__')

    def test__iadd__(self):

        expr = Expression(name='')

        func = expr.__iadd__(1)

        self._assertfunc(func, FuncName.IADD, expr, 1)
        self._asserttype(expr, '__iadd__')

    def test_iconcat(self):

        expr = Expression(name='')

        func = expr.__iadd__('')

        self._assertfunc(func, FuncName.ICONCAT, expr, '')
        self._asserttype(expr, '__iadd__')

    def test__sub__(self):

        expr = Expression(name='')

        func = expr - 1

        self._assertfunc(func, FuncName.SUB, expr, 1)
        self._asserttype(expr, '__sub__')

    def test__isub__(self):

        expr = Expression(name='')

        func = expr.__isub__(1)

        self._assertfunc(func, FuncName.ISUB, expr, 1)
        self._asserttype(expr, '__isub__')

    def test__mul__(self):

        expr = Expression(name='')

        func = expr * 1

        self._assertfunc(func, FuncName.MUL, expr, 1)
        self._asserttype(expr, '__mul__')

    def test__imul__(self):

        expr = Expression(name='')

        func = expr.__imul__(1)

        self._assertfunc(func, FuncName.IMUL, expr, 1)
        self._asserttype(expr, '__imul__')

    def test__truediv__(self):

        expr = Expression(name='')

        func = expr / 1

        self._assertfunc(func, FuncName.DIV, expr, 1)
        self._asserttype(expr, '__truediv__')

    def test__itruediv__(self):

        expr = Expression(name='')

        func = expr.__itruediv__(1)

        self._assertfunc(func, FuncName.IDIV, expr, 1)
        self._asserttype(expr, '__itruediv__')

    def test__invert__(self):

        expr = Expression(name='')

        func = ~expr

        self._assertfunc(func, FuncName.INVERT, expr)

    def test__neg__(self):

        expr = Expression(name='')

        func = -expr

        self._assertfunc(func, FuncName.NEG, expr)

    def test__abs__(self):

        expr = Expression(name='')

        func = abs(expr)

        self._assertfunc(func, FuncName.ABS, expr)

    def test__contains__(self):

        expr = Expression(name='')

        func = expr.__contains__('')

        self._assertfunc(func, FuncName.IN, expr, '')

    def test__mod__(self):

        expr = Expression(name='')

        func = expr % r''

        self._assertfunc(func, FuncName.LIKE, expr, '')
        self._asserttype(expr, '__mod__')

    def test__pow__(self):

        expr = Expression(name='')

        func = expr ** 1

        self._assertfunc(func, FuncName.POW, expr, 1)
        self._asserttype(expr, '__pow__')

    def test__ipow__(self):

        expr = Expression(name='')

        func = expr.__ipow__(1)

        self._assertfunc(func, FuncName.IPOW, expr, 1)
        self._asserttype(expr, '__ipow__')

    def test__radd__(self):

        expr = Expression(name='')

        func = 1 + expr

        self._assertfunc(func, FuncName.ADD, 1, expr)
        self._asserttype(expr, '__radd__')

    def test__rsub__(self):

        expr = Expression(name='')

        func = 1 - expr

        self._assertfunc(func, FuncName.SUB, 1, expr)
        self._asserttype(expr, '__rsub__')

    def test__rmul__(self):

        expr = Expression(name='')

        func = 1 * expr

        self._assertfunc(func, FuncName.MUL, 1, expr)
        self._asserttype(expr, '__rmul__')

    def test__rfloordiv__(self):

        expr = Expression(name='')

        func = 1 // expr

        self._assertfunc(func, FuncName.FLOORDIV, 1, expr)
        self._asserttype(expr, '__rfloordiv__')

    def test__rtruediv__(self):

        expr = Expression(name='')

        func = 1 / expr

        self._assertfunc(func, FuncName.DIV, 1, expr)
        self._asserttype(expr, '__rtruediv__')

    def test__rmod__(self):

        expr = Expression(name='')

        func = expr % expr

        self._assertfunc(func, FuncName.LIKE, expr, expr)
        self._asserttype(expr, '__rmod__')

    def test__rpow__(self):

        expr = Expression(name='')

        func = 1 ** expr

        self._assertfunc(func, FuncName.POW, 1, expr)
        self._asserttype(expr, '__rpow__')

    def test__getslice__(self):

        expr = Expression(name='')

        func = expr[1:2:3]

        self._assertfunc(func, FuncName.GETSLICE, expr, 1, 2, 3)
        self._asserttype(expr, '__getslice__')

    def test__setslice__(self):

        expr = Expression(name='')

        func = expr.__setslice__(1, 2, [])

        self._assertfunc(func, FuncName.SETSLICE, expr, 1, 2, [])
        self._asserttype(expr, '__setslice__')

    def test__delslice__(self):

        expr = Expression(name='')

        func = expr.__delslice__(1, 2)

        self._assertfunc(func, FuncName.DELSLICE, expr, 1, 2)
        self._asserttype(expr, '__delslice__')

    def test__getitem__(self):

        expr = Expression(name='')

        func = expr[1]

        self._assertfunc(func, FuncName.GETITEM, expr, 1)

    def test__setitem__(self):

        expr = Expression(name='')

        func = expr.__setitem__(1, True)

        self._assertfunc(func, FuncName.SETITEM, expr, 1, True)
        self._asserttype(expr, '__setitem__')

    def test__delitem__(self):

        expr = Expression(name='')

        func = expr.__delitem__(1)

        self._assertfunc(func, FuncName.DELITEM, expr, 1)

    def test__rshift__(self):

        expr = Expression(name='')

        func = expr >> 1

        self._assertfunc(func, FuncName.RSHIFT, expr, 1)

    def test__irshift__(self):

        expr = Expression(name='')

        func = expr.__irshift__(1)

        self._assertfunc(func, FuncName.IRSHIFT, expr, 1)

    def test__rrshift__(self):

        expr = Expression(name='')

        func = 1 >> expr

        self._assertfunc(func, FuncName.RSHIFT, 1, expr)

    def test__lshift__(self):

        expr = Expression(name='')

        func = expr << 1

        self._assertfunc(func, FuncName.LSHIFT, expr, 1)

    def test__ilshift__(self):

        expr = Expression(name='')

        func = expr.__ilshift__(1)

        self._assertfunc(func, FuncName.ILSHIFT, expr, 1)

    def test__rlshift__(self):

        expr = Expression(name='')

        func = 1 << expr

        self._assertfunc(func, FuncName.LSHIFT, 1, expr)

    def test_copy(self):

        expr = Expression(name='name', alias='alias')

        cexpr = expr.copy()

        self.assertIsNot(expr, cexpr)
        self.assertEqual(expr.name, cexpr.name)
        self.assertEqual(expr.alias, cexpr.alias)

    def test__call__(self):

        expr = Expression(name='test')
        params = (1,)
        func = expr(*params)

        self.assertIsInstance(func, Function)
        self.assertEqual(func.name, expr.name)
        self.assertEqual(func.params, params)


class FunctionTest(UTCase):

    def test_init_default(self):

        func = Function(name='')

        self.assertEqual(func.params, [])

    def test_init(self):

        params = [1, 2]

        func = Function(name='name', params=params)

        self.assertEqual(func.params, params)

    def test_call(self):

        params = [1, 2]

        func = Function(name='name')(1, 2)

        self.assertEqual(func.params, params)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""conf file driver UTs."""

from unittest import main

from b3j0f.utils.ut import UTCase

from ...expr import Expression as Exp
from ..create import Create


class CreateTest(UTCase):

    def test_init(self):

        values = {None: None}

        create = Create(name='test', values=values)

        self.assertEqual('test', create.name)
        self.assertEqual(values, create.values)

    def test___repr__(self):
        cases = [
            {
                'values': {},
                'name': Exp.c,
                'where': None,
                'as': None,
                'expected': "CREATE c:{}",
            },
            {
                'values': {Exp.key: 'value'},
                'name': Exp.c,
                'where': None,
                'as': None,
                'expected': "CREATE c:{key: 'value'}",
            },
            {
                'values': {Exp.k: 'v'},
                'name': '',
                'where': None,
                'as': None,
                'expected': "CREATE {k: 'v'}",
            },
            {
                'values': {Exp.k: 'v'},
                'name': '',
                'where': Exp.w,
                'as': None,
                'expected': "CREATE {k: 'v'} WHERE w",
            },
            {
                'values': {Exp.k: 'v'},
                'name': '',
                'where': None,
                'as': 'a',
                'expected': "CREATE {k: 'v'} AS a",
            },
        ]

        for test in cases:
            c = Create(
                values=test['values'],
                name=test['name'],
            )

            if test['where']:
                c = c.where(test['where'])

            if test['as']:
                c = c.as_(test['as'])

            self.assertEqual(repr(c), test['expected'])

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Context module."""

from collections import Hashable, Iterable

from six import iteritems

from ..request.base import BaseElement
from ..request.crud.join import Join, applyjoin

__all__ = ['Context', 'getctxname']


def getctxname(obj):
    """Get context name from input obj.

    :param obj: object from where get a context name.
    :return: obj context name.
    :rtype: str
    """
    result = obj

    if obj is not None:

        if isinstance(obj, BaseElement):
            result = obj.ctxname

        elif not isinstance(obj, Hashable):

            if isinstance(obj, Iterable):
                keys = sorted(list(obj))

                if isinstance(obj, dict):
                    _obj = [
                        (key, getctxname(obj[key])) for key in keys
                    ]

                else:
                    _obj = [getctxname(key) for key in keys]

                result = hash(str(_obj))

            else:
                result = id(obj)

    return result


class Context(dict):
    """Request execution context."""

    def __getitem__(self, key):

        ctxname = getctxname(key)

        return super(Context, self).__getitem__(ctxname)

    def __setitem__(self, key, value):

        ctxname = getctxname(key)

        return super(Context, self).__setitem__(ctxname, value)

    def __delitem__(self, key):

        ctxname = getctxname(key)

        return super(Context, self).__delitem__(ctxname)

    def __contains__(self, key):

        ctxname = getctxname(key)

        return super(Context, self).__contains__(ctxname)

    def get(self, key, default=None):

        ctxname = getctxname(key)

        return super(Context, self).get(ctxname, default)

    def pop(self, key, default=None):

        ctxname = getctxname(key)

        return super(Context, self).pop(ctxname, default)

    def setdefault(self, key, default):

        ctxname = getctxname(key)

        return super(Context, self).setdefault(ctxname, default)

    def fill(self, ctx, join=Join.FULL):
        """Fill this content with ctx data not in this data.

        :param Context ctx: ctx context from where get items."""

        if isinstance(ctx, Context):
            for key, value in iteritems(ctx):

                if key in list(self):
                    self[key] = applyjoin(self[key], value)

                else:
                    self[key] = ctx[key]

                dotkey = '.{0}'.format(key)

                for key in list(self):
                    if key.endswith(dotkey):
                        self[key] = applyjoin(self[key], value)

        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""request.base UTs."""

from collections import Hashable

from unittest import main

from b3j0f.utils.ut import UTCase

from ..ctx import Context, getctxname
from ...request.crud.base import BaseElement


class GetCtxNameTest(UTCase):

    def test_hashable(self):

        self.assertEqual(getctxname('test'), 'test')

    def test_baseelement(self):

        baseelement = BaseElement()

        self.assertEqual(getctxname(baseelement), baseelement.ctxname)

    def test_nothashable(self):

        elt = {}

        self.assertIsInstance(getctxname(elt), Hashable)

        self.assertEqual(getctxname(elt), getctxname(elt))

        self.assertEqual(getctxname({1: 1, 2: 2}), getctxname({2: 2, 1: 1}))

        self.assertNotEqual(getctxname({1: 1}), getctxname({1: None}))

    def test_nothashable_nothashable(self):

        elt = [[{1: []}]]

        self.assertIsInstance(getctxname(elt), Hashable)

        self.assertEqual(getctxname(elt), getctxname(list(elt)))


class ContextTest(UTCase):

    def test_init(self):

        self.assertFalse(Context())

    def test_init_params(self):

        context = Context({'a': 1})

        self.assertTrue(context)

        self.assertIn('a', context)

    def test_crud(self):

        context = Context()
        crud = BaseElement()

        self.assertNotIn(crud, context)

        context[crud] = 1

        self.assertIn(crud, context)
        self.assertIn(crud.ctxname, context)

        self.assertEqual(context[crud], 1)
        self.assertEqual(context[crud.ctxname], 1)

        del context[crud]

        self.assertNotIn(crud, context)
        self.assertNotIn(crud.ctxname, context)

    def test_fill(self):

        context = Context({
            'a.b': [1, 2, 4],
            'c': [5]
        })

        _context = Context({
            'b': [3],
            'c': [4],
            'a': [5]
        })

        context.fill(_context)

        self.assertEqual(context['a'], [5])
        self.assertEqual(context['b'], [3])
        self.assertEqual(context['c'], [5, 4])
        self.assertEqual(context['a.b'], [1, 2, 4, 3])

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""delete module."""

from six import iteritems

from .base import CRUDElement

__all__ = ['Delete']


class Delete(CRUDElement):

    __slots__ = ['_names'] + CRUDElement.__slots__

    def __init__(self, names=None, *args, **kwargs):
        """
        :param list names: model names to delete.
        """
        super(Delete, self).__init__(*args, **kwargs)

        self._names = () if names is None else names

    def names(self, *values):

        if values:
            self._names = values
            result = self

        else:
            result = self._names

        return result

    def __repr__(self):

        if self.names:
            names = [repr(name) for name in self._names]
            names = ', '.join(names)

        else:
            names = 'ALL'

        result = 'DELETE {0} '.format(names)

        if self.query:
            result += 'WHERE {0} '.format(repr(self.query))

        if self.dparams:
            result += 'WITH '

            dparams = []
            for name, value in iteritems(self.dparams):
                dparam = '{0}'.format(name)

                if value is not True:
                    dparam += ': {0}'.format(value)

                dparams.append(dparam)

            result += '{0} '.format(', '.join(dparams))

        if self.alias:
            result += 'AS {0}'.format(self.alias)

        if result[-1] == ' ':
            result = result[:-1]

        return result



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""Driver generator module."""

from six import iteritems

from inspect import getmembers, isroutine

from copy import deepcopy

from b3j0f.annotation import Annotation
from b3j0f.schema import Schema, data2schema

from .base import Driver
from .py import processread
from .transaction import State, Transaction
from .utils import getchildren
from ..request.consts import FuncName, CONDITIONS
from ..request.expr import Expression, Function
from ..request.crud.base import CRUD, CRUDElement
from ..request.crud.create import Create
from ..request.crud.read import Read
from ..request.crud.update import Update

try:
    from threading import Thread

except ImportError:
    from dummy_threading import Thread

__all__ = [
    'CustomDriver',
    'func2crudprocessing', 'obj2driver',
    'CreateAnnotation', 'ReadAnnotation', 'UpdateAnnotation',
    'DeleteAnnotation', 'query2kwargs'
]


class CustomDriver(Driver):
    """Driver with fine grained implementation of crud functions.

    This driver uses at most five set of functions for five respective crud
    types.

    The processing execute the right function for all request crud objects.

    Functions must takes in parameters a 'crud' object, a 'request' object and
    kwargs for specific driver uses (like explain for example).

    Functions must return a request."""

    def __init__(
            self, creates=None, reads=None, updates=None, deletes=None,
            functions=None, *args, **kwargs
    ):
        """
        :param list creates: creation functions. Take in parameters a request,
            a crud operation and specific kwargs. Return created item
            with additional fields such as id.
        :param list reads: reading functions. Take in parameters a request,
            a crud operation and specific kwargs. Return read items.
        :param list updates: updating functions. Take in parameters a request,
            a crud operation and specific kwargs. Return updated items.
        :param list deletes: deletion functions. Take in parameters a request,
            a crud operation and specific kwargs. Return deleted items.
        :param dict functions: function to process for specific query functions
        .
        """

        super(CustomDriver, self).__init__(*args, **kwargs)

        self.creates = [] if creates is None else creates
        self.reads = [] if reads is None else reads
        self.updates = [] if updates is None else updates
        self.deletes = [] if deletes is None else deletes
        self.functions = {} if functions is None else functions

    def _process(self, transaction, **kwargs):

        result = transaction

        if transaction.state is State.COMMITTING:

            for crud in transaction.cruds:

                self.processquery(
                    query=crud.query, transaction=transaction, **kwargs
                )

                crudname = type(crud).__name__.lower()
                funcs = getattr(self, '{0}s'.format(crudname))

                if funcs:
                    transaction.ctx[crud] = []

                    for func in funcs:
                        fresult = func(crud=crud, transaction=result, **kwargs)

                        if isinstance(fresult, list):
                            transaction.ctx[crud] += fresult

                        elif isinstance(fresult, Transaction):
                            transaction.ctx[crud] += fresult.ctx[crud]

                        elif fresult is not None:
                            msg1 = 'Wrong func result {0}.'.format(func)
                            msg2 = 'Transaction, list or None expected.'
                            raise TypeError('{0} {1}'.format(msg1, msg2))

                else:
                    raise NotImplementedError(
                        'No implementation found for {0}'.format(crudname)
                    )

        return result

    def processquery(self, query, transaction, **kwargs):

        result = query

        children = getchildren(query)

        for child in children:
            self.processquery(query=child, transaction=transaction, **kwargs)

        if isinstance(query, Function):

            if query.name in self.supportedfunctions:
                pass

            elif query.name in self.functions:

                func = self.functions[query.name]
                result = func(
                    query=query, transaction=transaction, **kwargs
                )
                transaction.ctx[query] = result

            else:
                raise NotImplementedError(
                    '{0} is not implemented by {1}'.format(query, self)
                )

        return result


def func2crudprocessing(func, annotation):

    if not isinstance(func, Schema):
        func = data2schema(func)

    def _processing(crud, transaction, _func=func, **kwargs):

        if annotation.gateway:

            pnames = [param.name for param in _func.params]

            orfunckwargs = query2kwargs(
                query=crud.query, ctx=transaction.ctx, pnames=pnames
            )

            funcresult = []

            threads = []

            for funckwargs in orfunckwargs:

                funcvarargs = []

                for param in _func.params:

                    pname = param.name

                    if pname in transaction.ctx and pname not in funckwargs:

                        funckwargs[param.name] = transaction.ctx[param.name]

                if annotation.translator is not None:
                    for key in list(funckwargs):
                        val = funckwargs[key]

                        funckwargs[key] = annotation.translator(key, val)

                if isinstance(crud, (Create, Update)):

                    for key in crud.values:
                        if key in pnames:
                            funckwargs[key] = crud.values[key]

                if len(orfunckwargs) == 1:

                    try:
                        funcresult = _func(*funcvarargs, **funckwargs)

                    except Exception:
                        pass

                else:
                    def target(funcresult=funcresult):
                        try:
                            result = _func(*funcvarargs, **funckwargs)

                        except Exception:
                            pass

                        else:
                            funcresult += result

                    thread = Thread(target=target)
                    thread.start()
                    threads.append(thread)

            # TODO: remove duplicate data
            for thread in threads:
                thread.join()

            transaction.ctx[crud] = [] if funcresult is None else funcresult

            if isinstance(crud, Read):
                funcresult = processread(
                    read=crud,
                    items=funcresult,
                    ctx=transaction.ctx
                )

        else:

            fresult = transaction

            try:
                fresult = _func(transaction=transaction, crud=crud, **kwargs)

            except Exception:
                pass

            if fresult is None:

                funcresult = transaction.ctx[crud]

            else:
                funcresult = fresult.ctx[crud]

        transaction.ctx[crud] = funcresult

        return transaction

    return _processing


def query2kwargs(query, ctx, pnames):
    """Parse input query and return a list of kwargs for crud function.

    The list indicates

    :param Expression query: query to parse in order to get conditions.
    :param Context ctx: execution context.
    :param list pnames: parameter names to identify such as conditional
        expressions from the query.
    :return: list of kwargs where keys are parameter names and values are dict
        of tuple of operator name, parameters.
    :rtype: list
    """
    result = [{}]

    if isinstance(query, Expression):

        if query.name in pnames:

            for item in result:
                item[query.name] = ctx.get(query, {FuncName.EXISTS.value: []})

        if isinstance(query, Function):

            if query.name in CONDITIONS:

                isor = query.name == FuncName.OR.value

                if isor:
                    orresult = result
                    result = []

                if isor or query.name == FuncName.AND.value:

                    for param in query.params:

                        queryresult = query2kwargs(
                            query=param, ctx=ctx, pnames=pnames
                        )

                        fresult = deepcopy(orresult) if isor else result

                        for item in fresult:

                            for qitem in queryresult:

                                for qname, qconds in iteritems(qitem):

                                    if qname in item:
                                        item[qname].update(qconds)

                                    else:
                                        item[qname] = qconds

                        if isor:
                            result += fresult

                        else:
                            result = fresult

                elif query.params:
                    elt = query.params[0]

                    if not isinstance(elt, Expression):
                        raise TypeError(
                            'Wrong type {0} in {1}. Expr expected.'.format(
                                elt, query
                            )
                        )

                    if elt.name in pnames:

                        params = [
                            ctx.get(param, param) for param in query.params[1:]
                        ]

                        for item in result:
                            item.setdefault(elt.name, {})[query.name] = params

                else:
                    raise NotImplementedError(
                        'Parameters are missing on {0}.'.format(query)
                    )

    elif isinstance(query, CRUDElement):

        if query.ctxname in pnames:

            for item in result:
                item[query.ctxname] = ctx.get(query, query)

    return result


def obj2driver(
        obj, name=None,
        creates=None, reads=None, updates=None, deletes=None, functions=None
):
    """Convert an object to a driver.

    :param str name: driver name. obj type name by default.
    :param list creates: create function names to retrieve from the obj.
    :param list reads: read function names to retrieve from the obj.
    :param list updates: update function names to retrieve from the obj.
    :param list deletes: delete function names to retrieve from the obj.
    :param dict functions: fuctions by name. obj functions by default.
    :rtype: CustomDriver
    """

    if name is None:
        name = type(obj).__name__

    # build obj schema if necessary
    if not isinstance(obj, Schema):
        fobj = data2schema(obj, _force=True)

    else:
        fobj = obj

    if creates is None:
        creates = []

    if reads is None:
        reads = []

    if updates is None:
        updates = []

    if deletes is None:
        deletes = []

    if functions is None:
        functions = {}

    _locals = locals()

    # get data from crud annotations
    crudannotations = _CRUDAnnotation.get_annotations(obj, maxdepth=2)
    for crudannotation in crudannotations:

        for target in crudannotation.targets:

            otarget = getattr(obj, target.__name__)

            ftarget = func2crudprocessing(otarget, crudannotation)

            _locals['{0}s'.format(crudannotation.name)].append(ftarget)

    for name, member in getmembers(fobj):

        if isroutine(member):

            functions[name] = member

    return CustomDriver(
        name=name,
        creates=creates,
        reads=reads,
        updates=updates,
        deletes=deletes,
        functions=functions
    )


class _CRUDAnnotation(Annotation):
    """Base Annotation for annotating CRUD functions."""

    def __init__(self, crud, gateway=True, translator=None, *args, **kwargs):
        """
        :param str crud: crud name.
        :param bool gateway: pass arguments from the context to the function.
        :param callable translator: if gateway, translate gateway parameters to
            translated data. Must take a gateway parameter name and value in
            parameters. For example you could use `datafromgateway`.
        """

        super(_CRUDAnnotation, self).__init__(*args, **kwargs)

        if isinstance(crud, CRUD):
            crud = crud.name.lower()

        self.name = crud
        self.gateway = gateway
        self.translator = translator


class CreateAnnotation(_CRUDAnnotation):
    """Annotate create functions."""

    def __init__(self, *args, **kwargs):

        super(CreateAnnotation, self).__init__(
            crud=CRUD.CREATE, *args, **kwargs
        )


class ReadAnnotation(_CRUDAnnotation):
    """Annotate read functions."""

    def __init__(self, *args, **kwargs):

        super(ReadAnnotation, self).__init__(
            crud=CRUD.READ, *args, **kwargs
        )


class UpdateAnnotation(_CRUDAnnotation):
    """Annotate update functions."""

    def __init__(self, *args, **kwargs):

        super(UpdateAnnotation, self).__init__(
            crud=CRUD.UPDATE, *args, **kwargs
        )


class DeleteAnnotation(_CRUDAnnotation):
    """Annotate delete functions."""

    def __init__(self, *args, **kwargs):

        super(DeleteAnnotation, self).__init__(
            crud=CRUD.DELETE, *args, **kwargs
        )


def datafromgateway(name, param):
    """Get a data from parameter."""

    result = param

    if isinstance(param, dict):
        for key, value in iteritems(param):
            if key == FuncName.EQ.value:
                result = value[0]

            elif key == FuncName.BETWEEN.value:
                result = value

            elif key == FuncName.IN.value:
                result = value

    return result



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""join execution module."""

from copy import deepcopy

from enum import IntEnum, unique

__all__ = [
    'innerjoin', 'leftjoin', 'leftexjoin', 'rightjoin', 'rightexjoin',
    'fulljoin', 'fullexjoin', 'crossjoin', 'selfjoin', 'naturaljoin',
    'unionjoin', 'applyjoin'
]


@unique
class Join(IntEnum):

    INNER = 0  #: inner join.
    LEFT = 1  #: left join.
    LEFTEX = 2  #: left exclusive join.
    RIGHT = 3  #: right exclusive join.
    RIGHTEX = 4  #: right exclusive join.
    FULL = 5  #: full join.
    FULLEX = 6  #: full exclusive join.
    CROSS = 7  #: cross join.
    SELF = 8  #: self join.
    NATURAL = 9  #: natural join.
    UNION = 10  #: union join.


def innerjoin(litems, ritems):

    return [item for item in litems if item in ritems]


def leftjoin(litems, ritems):

    return litems


def leftexjoin(litems, ritems):

    return [item for item in litems if item not in ritems]


def rightjoin(litems, ritems):

    return ritems


def rightexjoin(litems, ritems):

    return [item for item in ritems if item not in litems]


def fulljoin(litems, ritems):
    """Apply full join on litems and rtimes.

    :param list litems:
    :param list ritmes:
    :return: new list of items.
    :rtype: list"""

    return litems + [item for item in ritems if item not in litems]


def fullexjoin(litems, ritems):

    return leftexjoin(litems, ritems) + rightexjoin(litems, ritems)


def crossjoin(litems, ritems):

    return [(litem, ritem) for litem in litems for ritem in ritems]


def selfjoin(litems, ritems):

    return crossjoin(litems, litems)


def naturaljoin(litems, ritems):

    result = []

    for litem in litems:

        for ritem in ritems:

            issame = False

            for key in ritem:

                if key in litem:
                    if litem[key] == ritem[key]:
                        issame = True

                    else:
                        break

            else:
                if issame:
                    item = deepcopy(litem)
                    item.update(deepcopy(ritem))
                    result.append(item)

    return result


def unionjoin(litems, ritems):

    return litems + ritems


_JOINBYNAME = {
    Join.INNER.name: innerjoin,
    Join.LEFT.name: leftjoin,
    Join.LEFTEX.name: leftexjoin,
    Join.RIGHT.name: rightjoin,
    Join.RIGHTEX.name: rightexjoin,
    Join.FULL.name: fulljoin,
    Join.FULLEX.name: fullexjoin,
    Join.CROSS.name: crossjoin,
    Join.SELF.name: selfjoin,
    Join.NATURAL.name: naturaljoin,
    Join.UNION.name: unionjoin
}


def applyjoin(litems, ritems, join=Join.FULL):

    if isinstance(join, Join):
        join = join.name

    func = _JOINBYNAME[join]

    return func(litems, ritems)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2016 Jonathan Labéjof <jonathan.labejof@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# --------------------------------------------------------------------

"""conf file driver UTs."""

from unittest import main

from b3j0f.utils.ut import UTCase

from ..ctx import Context
from ..py import (
    PyDriver,
    getsubitem,
    processcreate, processcrud, processdelete, processquery, processread,
    processupdate
)
from ...request.crud.create import Create
from ...request.crud.delete import Delete
from ...request.crud.read import Read
from ...request.crud.update import Update
from ...request.expr import Expression, Function


class CRUDTest(UTCase):

    def setUp(self):

        self.items = [
            {'name': str(i), 'id': i} for i in range(5)
        ]

        self.items[-1]['ext'] = None
        self.items[-1]['name'] = str(0)


class CreateTest(CRUDTest):

    def test(self):

        crud = Create(name='test', values={})

        result = processcreate(items=self.items, create=crud)

        self.assertIn(crud.values, self.items)

        self.assertIs(result, self.items)


class ReadTest(CRUDTest):

    def test_default(self):

        crud = Read()

        result = processread(items=self.items, read=crud)

        self.assertEqual(result, self.items)

        self.assertIsNot(result, self.items)

    def test_select(self):

        crud = Read(select=['id'])

        result = processread(items=self.items, read=crud)

        self.assertEqual(result, [{'id': i} for i in range(5)])

        self.assertIsNot(result, self.items)

    def test_offset(self):

        crud = Read(offset=1)

        result = processread(items=self.items, read=crud)

        self.assertEqual(result, self.items[1:])

        self.assertIsNot(result, self.items)

    def test_limit(self):

        crud = Read(limit=1)

        result = processread(items=self.items, read=crud)

        self.assertEqual(result, self.items[:1])

        self.assertIsNot(result, self.items)

    def test_groupby(self):

        crud = Read(groupby='name')

        result = processread(items=self.items, read=crud)

        self.assertEqual(
            result,
            [
                {'id': [1], 'name': '1'},
                {'ext': [None], 'id': [0, 4], 'name': '0'},
                {'id': [3], 'name': '3'},
                {'id': [2], 'name': '2'}
            ]
        )

    def test_join(self):

        crud = Read(join='')

        self.assertRaises(
            NotImplementedError, processread, items=self.items, read=crud
        )


class UpdateTest(CRUDTest):

    def test(self):

        crud = Update(name='', values={'name': 1})

        result = processupdate(items=self.items, update=crud)

        self.assertIs(result, self.items)

        self.assertEqual(result[:4], [{'name': 1, 'id': i} for i in range(4)])
        self.assertEqual(result[4], {'name': 1, 'id': 4, 'ext': None})


class DeleteTest(CRUDTest):

    def test_default(self):

        crud = Delete()

        result = processdelete(items=self.items, delete=crud)

        self.assertIs(result, self.items)

        self.assertFalse(result)

    def test_name(self):

        crud = Delete(names=('name', ))

        result = processdelete(items=self.items, delete=crud)

        self.assertIs(result, self.items)

        self.assertEqual(result[:4], [{'id': i} for i in range(4)])
        self.assertEqual(result[4], {'id': 4, 'ext': None})


class ProcessCRUDTest(CRUDTest):

    def setUp(self):

        super(ProcessCRUDTest, self).setUp()

        self.ctx = Context()

    def _assert(self, crud):

        result = processcrud(crud=crud, ctx=self.ctx, items=self.items)

        self.assertEqual(result, self.ctx[crud])

    def test_create(self):

        crud = Create(name='', values={})

        self._assert(crud=crud)

    def test_read(self):

        crud = Read()

        self._assert(crud=crud)

    def test_update(self):

        crud = Update(name='', values={})

        self._assert(crud=crud)

    def test_delete(self):

        crud = Delete()

        self._assert(crud=crud)


class ProcessQueryTest(CRUDTest):

    def test_expr(self):

        result = processquery(items=self.items, query=Expression.ext)

        self.assertEqual(result, [self.items[-1]])

    def test_lookup(self):

        result = processquery(
            items=self.items, query=Expression.b3j0f.utils.ut.UTCase
        )

        self.assertIs(result, UTCase)
        self.assertTrue(self.items)

    def test_is(self):

        result = processquery(
            items=self.items, query=Function.is_(Expression.id, 2)
        )

        self.assertEqual(result, [self.items[2]])

    def test_isnot(self):

        result = processquery(
            items=self.items, query=Function.isnot(Expression.id, 2)
        )

        self.assertEqual(result, self.items[:2] + self.items[3:])

    def test_eq(self):

        result = processquery(items=self.items, query=Expression.name_ == '2')

        self.assertEqual(result, [self.items[2]])

    def test_ne(self):

        result = processquery(items=self.items, query=Expression.id != 0)

        self.assertEqual(result, self.items[1:])

    def test_gt(self):

        result = processquery(items=self.items, query=Expression.id > 2)

        self.assertEqual(result, self.items[3:])

    def test_ge(self):

        result = processquery(items=self.items, query=Expression.id >= 2)

        self.assertEqual(result, self.items[2:])

    def test_lt(self):

        result = processquery(items=self.items, query=Expression.id < 2)

        self.assertEqual(result, self.items[:2])

    def test_le(self):

        result = processquery(items=self.items, query=Expression.id <= 2)

        self.assertEqual(result, self.items[:3])

    def test_like(self):

        result = processquery(
            items=self.items, query=Expression.name_ % '[1234]'
        )

        self.assertEqual(result, self.items[1:-1])

    def test_exists(self):

        result = processquery(
            items=self.items, query=Expression.exists(Expression.ext)
        )

        self.assertEqual(result, [self.items[-1]])

    def test_nexists(self):

        result = processquery(
            items=self.items, query=Expression.nexists(Expression.ext)
        )

        self.assertEqual(result, self.items[:-1])

    def test_isnull(self):

        result = processquery(
            items=self.items, query=Expression.isnull(Expression.ext)
        )

        self.assertEqual(result, [self.items[-1]])

    def test_between(self):

        result = processquery(
            items=self.items, query=Expression.between(Expression.id, 1, 2)
        )

        self.assertEqual(result, self.items[1:3])

    def test_in(self):

        result = processquery(
            items=self.items, query=Expression.in_(Expression.id, [1, 2])
        )

        self.assertEqual(result, self.items[1:3])

    def test_having(self):

        result = processquery(
            items=self.items, query=Expression.having(
                Expression.in_(Expression.id, [1, 2])
            )
        )

        self.assertEqual(result, self.items[1:3])

    def test_all(self):

        result = processquery(
            items=self.items, query=Expression.all(
                Expression.id, Function('>'), [1, 2]
            )
        )

        self.assertEqual(result, self.items[3:])

        result = processquery(
            items=self.items, query=Expression.all(Expression.id, '>', [1, 2])
        )

        self.assertEqual(result, self.items[3:])

    def test_any(self):

        result = processquery(
            items=self.items,
            query=Expression.any(Expression.id, Function('>'), [1, 2])
        )

        self.assertEqual(result, self.items[2:])

        result = processquery(
            items=self.items,
            query=Expression.any(Expression.id, '>', [1, 2])
        )

        self.assertEqual(result, self.items[2:])

    def test_some(self):

        result = processquery(
            items=self.items,
            query=Expression.some(Expression.id, Function('>'), [1, 2])
        )

        self.assertEqual(result, self.items[2:])

        result = processquery(
            items=self.items,
            query=Expression.some(Expression.id, '>', [1, 2])
        )

        self.assertEqual(result, self.items[2:])

    def test_or(self):

        result = processquery(
            items=self.items,
            query=(Expression.id < 2) | (Expression.id > 3)
        )

        self.assertEqual(result, self.items[:2] + self.items[4:])

    def test_and(self):

        result = processquery(
            items=self.items, query=(Expression.id > 2) & (Expression.id < 4)
        )

        self.assertEqual(result, [self.items[3]])


class GetSubItemTest(UTCase):

    def setUp(self):

        self.item = {'test': {'test': {'test': None}}}

    def test_one(self):

        subitem = getsubitem(item=self.item, name='test')

        self.assertIs(subitem, self.item['test'])

    def test_two(self):

        subitem = getsubitem(item=self.item, name='test.test')

        self.assertIs(subitem, self.item['test']['test'])

    def test_falseerror(self):

        subitem = getsubitem(item=self.item, name='a')

        self.assertIsNone(subitem)

    def test_error(self):

        self.assertRaises(
            KeyError, getsubitem, item=self.item, name='a', error=True
        )


class PyDriverTest(UTCase):

    def setUp(self):

        self.driver = PyDriver()

if __name__ == '__main__':
    main()
